import hashlib
import json
import uuid
from dataclasses import dataclass
from datetime import datetime
from typing import Any

import pydash
from adafri.cache.lib import CacheClass
from adafri.v1.files import (
    IMAGES_EXTENSIONS,
    File,
    crop_image,
    crop_image_from_url,
    estimate_cropped_file_size,
    estimate_pre_crop_file_size,
    get_file_size_from_url,
    upload_image_from_url,
)
from adafri.v1.gads.data import NATIVE_ADS_REQUIREMENTS as IMAGES_RATIO
from adafri.v1.gads.models.client_customer_id import ClientCustomerId
from firebase_admin.firestore import firestore

from app.test import cache

from ......utils import (
    ArrayUtils,
    BaseClass,
    DateUtils,
    DictUtils,
    get_object_model_class,
    get_request_fields,
    init_class_kwargs,
    integer,
    string,
)
from ......utils.response import ApiResponse, Error, ResponseStatus, StatusCode
from ......v1 import version
from .....account import Account
from .....base.firebase_collection import FirebaseCollectionBase
from ....models.client_customer_id import ClientCustomerId
from .asset_fields import (
    ASSET_CROP_FIELDS,
    ASSETS_COLLECTION,
    ASSETS_PICKABLE_FIELDS,
    FULL_SIZE_INFO_FIELDS,
    STANDARD_FIELDS,
    AssetCropFields,
    AssetCropFieldsProps,
    AssetFullSizeInfoFields,
    AssetFullSizeInfoFieldsProps,
    AssetsFields,
    AssetsFieldsProps,
)

RESPONSIVE_DISPLAY_ADS_UPLOAD_COLLECTION = "responsive_display_ads_upload"


def filter_asset_request_fields(fields, default_fields=ASSETS_PICKABLE_FIELDS):
    request_fields = get_request_fields(fields, default_fields, [default_fields[0]])
    if request_fields is None or bool(request_fields) is False:
        request_fields = [default_fields[0]]
    return request_fields


@dataclass(init=False)
class AssetFullSizeInfo(BaseClass):
    imageUrl: str | None = None
    imageWidth: int | None = None
    imageHeight: int | None = None

    def __init__(self, fullSizeInfo=None, **kwargs):
        (cls_object, keys, data_args) = init_class_kwargs(
            self,
            fullSizeInfo,
            FULL_SIZE_INFO_FIELDS,
            AssetFullSizeInfoFieldsProps,
            None,
            [],
            **kwargs,
        )
        super().__init__(**data_args)
        for key in keys:
            if key == AssetFullSizeInfoFields.imageUrl:
                setattr(self, key, string(cls_object[key], None, True))
            elif key == AssetFullSizeInfoFields.imageWidth:
                setattr(self, key, integer(cls_object[key]))
            elif key == AssetFullSizeInfoFields.imageHeight:
                setattr(self, key, integer(cls_object[key]))
            else:
                setattr(self, key, cls_object[key])

    @staticmethod
    def from_dict(asset: Any) -> "AssetFullSizeInfo":
        cls_object, keys = get_object_model_class(
            asset, AssetFullSizeInfo, AssetsFieldsProps
        )
        _asset = AssetFullSizeInfo(cls_object)
        return _asset

    def to_dict(self, _fields=None) -> "dict | None":
        return DictUtils.remove_none_values(
            {
                "imageUrl": self.imageUrl,
                "imageWidth": self.imageWidth,
                "imageHeight": self.imageHeight,
            }
        )

    def to_json(self, _fields=None):
        return self.to_dict(_fields)


class AssetCrop(BaseClass):
    x: int = None
    y: int = None
    xPosition: int = None
    yPosition: int = None

    def __init__(self, crop=None, **kwargs):
        (cls_object, keys, data_args) = init_class_kwargs(
            self, crop, ASSET_CROP_FIELDS, AssetCropFieldsProps, None, [], **kwargs
        )
        super().__init__(**data_args)
        for key in keys:
            if key == AssetCropFields.x:
                setattr(self, key, integer(cls_object[key], None, True))
            elif key == AssetCropFields.y:
                setattr(self, key, integer(cls_object[key]))
            elif key == AssetCropFields.xPosition:
                setattr(self, key, integer(cls_object[key]))
            elif key == AssetCropFields.yPosition:
                setattr(self, key, integer(cls_object[key]))
            else:
                setattr(self, key, cls_object[key])

    @staticmethod
    def from_dict(asset: Any) -> "AssetCrop":
        cls_object, keys = get_object_model_class(
            asset, AssetCrop, AssetCropFieldsProps
        )
        _asset = AssetCrop(cls_object)
        return _asset

    def to_dict(self, _fields=None):
        return DictUtils.remove_none_values(
            {
                "x": self.x,
                "y": self.y,
                "xPosition": self.xPosition,
                "yPosition": self.yPosition,
            }
        )


@dataclass(init=False)
class Asset(FirebaseCollectionBase):
    id: str = None
    hash: str = None
    name: str = None
    accountId: str = None
    owner: str = None
    clientCustomerId: str = None
    assetId: int = None
    imageMimeType: str = None
    imageFileSize: int = None
    assetName: str = None
    assetSubType: str = None
    fullSizeInfo: AssetFullSizeInfo = None
    usage: list[str] = None
    useFor: str = None
    width: int = None
    height: int = None
    imageUrl: str = None
    url: str = None
    originalImageInfo: str = None
    crop: AssetCrop = None
    crop_state: str = None
    content_type: str = None
    data: str = None
    size: int = None
    file_url: str = None
    type: str = None
    state: str = None
    account: Account | None = None
    customer: ClientCustomerId | None = None
    adgroup: Any | None = None
    __baseFields = AssetsFieldsProps

    def __init__(self, asset=None, **kwargs):
        _asset = asset
        if type(asset) is str:
            _asset = {"id": asset}
        if type(asset) is dict:
            _asset = _asset.copy()
            if "aacid" in _asset and "accountId" not in _asset:
                _asset["accountId"] = _asset["aacid"]
            if "url" in _asset and "imageUrl" not in _asset:
                _asset["imageUrl"] = _asset["url"]
            if "originalImageInfo" in _asset and type(asset) is dict:
                info = _asset.copy()["originalImageInfo"]
                if isinstance(info, dict) is True:
                    _asset["originalImageInfo"] = json.dumps(info)
                else:
                    _asset["originalImageInfo"] = info

        (cls_object, keys, data_args) = init_class_kwargs(
            self,
            _asset,
            STANDARD_FIELDS,
            AssetsFieldsProps,
            ASSETS_COLLECTION,
            ["id"],
            **kwargs,
        )
        super().__init__(**data_args)
        for key in keys:
            if key == AssetsFields.assetId:
                setattr(self, key, integer(cls_object[key]))
            elif key == AssetsFields.size:
                setattr(self, key, integer(cls_object[key]))
            elif key == AssetsFields.imageFileSize:
                setattr(self, key, integer(cls_object[key]))
            elif key == AssetsFields.width:
                setattr(self, key, integer(cls_object[key]))
            elif key == AssetsFields.height:
                setattr(self, key, integer(cls_object[key]))
            elif key == AssetsFields.fullSizeInfo:
                if bool(cls_object[key]) is True:
                    setattr(self, key, AssetFullSizeInfo.from_dict(cls_object[key]))
                else:
                    setattr(self, key, None)
            elif key == AssetsFields.imageMimeType:
                setattr(self, key, string(cls_object[key], None, True))
            elif key == AssetsFields.usage:
                setattr(self, key, ArrayUtils.to_array(cls_object[key]))
            elif key == AssetsFields.useFor:
                setattr(self, key, string(cls_object[key], None, True))
            elif key == AssetsFields.file_url:
                setattr(self, key, string(cls_object[key], None, True))
            elif key == AssetsFields.type:
                setattr(self, key, string(cls_object[key], None, True))
            elif key == AssetsFields.imageUrl:
                setattr(self, key, string(cls_object[key], None, True))
            elif key == AssetsFields.originalImageInfo:
                setattr(self, key, string(cls_object[key], None, True))
            elif key == AssetsFields.clientCustomerId:
                setattr(self, key, string(cls_object[key]))
            elif key == AssetsFields.crop_state:
                setattr(self, key, string(cls_object[key], None, True))
            elif key == AssetsFields.content_type:
                setattr(self, key, string(cls_object[key], None, True))
            elif key == AssetsFields.state:
                setattr(self, key, string(cls_object[key], None, True))
            elif key == AssetsFields.hash:
                setattr(self, key, string(cls_object[key], None, True))
            elif key == AssetsFields.account:
                setattr(self, key, Account.from_dict(cls_object[key]))
            elif key == AssetsFields.customer:
                setattr(self, key, ClientCustomerId.from_dict(cls_object[key]))
            elif key == AssetsFields.adgroup:
                setattr(self, key, cls_object[key])
            elif key == AssetsFields.crop and bool(cls_object[key]) is True:
                setattr(self, key, AssetCrop.from_dict(cls_object[key]))
            else:
                setattr(self, key, cls_object[key])
        self.__normalize__()
        self.cache = CacheClass("asset", Asset)


    def __normalize__(self):
        if bool(self.fullSizeInfo) is True and bool(self.fullSizeInfo.imageUrl) is True:
            if bool(self.imageUrl) is False:
                self.imageUrl = self.fullSizeInfo.imageUrl
                self.width = self.fullSizeInfo.imageWidth
                self.height = self.fullSizeInfo.imageHeight
        if bool(self.url) is True and bool(self.imageUrl) is False:
            self.imageUrl = self.url
        if bool(self.imageUrl) is True and bool(self.url) is False:
            self.url = self.imageUrl
        if bool(self.file_url) is False and bool(self.imageUrl) is True:
            self.file_url = self.imageUrl
    @staticmethod
    def generate_model(_key_="default_value"):
        user = {}
        props = AssetsFieldsProps
        for k in DictUtils.get_keys(props):
            user[k] = props[k][_key_]
        return user

    @staticmethod
    def from_dict(asset: Any, db=None, collection_name=None) -> "Asset":
        # print('asset', asset)
        _asset = None
        if type(asset) is dict:
            _asset = asset.copy()
            if "aacid" in _asset and "accountId" not in _asset:
                _asset["accountId"] = _asset["aacid"]
            if "url" in _asset and "imageUrl" not in _asset:
                _asset["imageUrl"] = _asset["url"]
            if "originalImageInfo" in _asset and type(asset) is dict:
                info = _asset.copy()["originalImageInfo"]
                if isinstance(info, dict) is True:
                    # print('info dict')
                    _asset["originalImageInfo"] = json.dumps(info)
                else:
                    _asset["originalImageInfo"] = info
        else:
            _asset = asset

        cls_object, keys = get_object_model_class(_asset, Asset, AssetsFieldsProps)
        _asset = Asset(cls_object, db=db, collection_name=collection_name)
        if None in _asset.usage:
            print('asset none', cls_object)
        return _asset
    def getAssetNative(self, id: str):
        db = FirebaseCollectionBase(collection_name='assets-native')
        doc = db.document_reference(id).get()
        if doc.exists is True:
            return doc.to_dict()
        return None
    def check_parsable(self, element: str):
        obj = None;
        try:
            obj = json.loads(element);
            keys = ['x', 'y', 'xPosition', 'yPosition']
            for key in keys:
                if key not in obj:
                    return {"is_url": False, "parsed": None}
            return {"is_url": False, "parsed": obj}
        except Exception:
            # print('ex', e)
            return {"is_url": True, "parsed": None, "url": element}
    def crop_image(self, element, folder: str):
        crop = None;
        if 'crop' in element:
            crop = element['crop'];
            crop['yPosition'] = crop['y'] + crop['yPosition'];
            crop['xPosition'] = crop['x'] + crop['xPosition'];
        else:
            crop = {
                'x': element['x'],
                'y': element['y'],
                'xPosition': element['xPosition'],
                'yPosition': element['yPosition']
            }
        data_id = None
        data_url = None
        if 'id' in element:
            data_id = element['id'];
        if 'originalImage' in element:
            data_url = element['originalImage'];
        elif 'originalImageInfo' in element:
            data_url = element['originalImageInfo']['url'];
        else:
            # data['url'] = element['url'];
            data_url = element['url'];
        cropping = crop_image_from_url(data_url, self.accountId, crop, folder, 'adafri_upload', data_id);
        if cropping is None:
            return None
        return cropping
    def createAssetFile(self):
        file = None
        if bool(self.hash) is False:
            if bool(self.url) is True:
                try:
                    file = upload_image_from_url(file=self.url, upload_path='', extensions=IMAGES_EXTENSIONS.copy(),accountId=self.accountId, upload_folder=RESPONSIVE_DISPLAY_ADS_UPLOAD_COLLECTION, process=False)
                    if file is not None:
                        return ApiResponse(ResponseStatus.OK, StatusCode.status_200, file, None)
                except:
                    pass
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"Cannot build asset file ref for {self.id}", "INVALID_REQUEST", 1))
        file = File({"id": self.hash}, collection_name=RESPONSIVE_DISPLAY_ADS_UPLOAD_COLLECTION).get()
        if file is not None:
            return ApiResponse(ResponseStatus.OK, StatusCode.status_200, file, None)
        return ApiResponse(ResponseStatus.OK, StatusCode.status_204, None, None)

         
    def build_parsable(self, file=None):
        _file = file if file is not None else File(file={"id": self.hash}, collection_name=RESPONSIVE_DISPLAY_ADS_UPLOAD_COLLECTION).get()
        parsable = self.check_parsable(self.url)
        if parsable is not None:
            if parsable['is_url'] is True and parsable['parsed'] is None:
                if _file is None:
                    # return ApiResponse(ResponseStatus.ERROR, StatusCode.status_500, None, Error(f"Need to create file {self.imageUrl} {self.hash}", "INVALID_REQUEST", 1))
                    asset_create = self.createAssetFile()
                    if asset_create.status == ResponseStatus.OK:
                        if asset_create.status_code == StatusCode.status_200:
                            return asset_create;
                    return ApiResponse(ResponseStatus.ERROR, StatusCode.status_500, None, Error(f"Cannot find or create asset file to use {self.id}", "INVALID_REQUEST", 1))
                return ApiResponse(ResponseStatus.OK, StatusCode.status_200, _file, None)
            
            # if parsable['is_url'] is False and parsable['parsed'] is not None:
            #     crop = self.crop_image(parsable['parsed'], 'assets-native')
            #     if crop is not None:
            #         self.file_url = crop.file_url
            #         self.width = crop.width
            #         self.height = crop.height
            #         eligibility = self.check_eligibility(file)
            #         if eligibility.status == ResponseStatus.ERROR:
            #             return eligibility
            #         self.crop_state = "done"
            #         return ApiResponse(ResponseStatus.OK, StatusCode.status_200, Asset.from_dict(self.to_dict()), None)  
        return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"Cannot build asset from parsed elements {self.id}", "INVALID_REQUEST", 1))           
    
    
    def build_asset(self, file: File | None=None, cropAsset = True, fetch = True, createAsset = True, test = False):
        asset_data = self;
        if fetch is True and bool(self.id) is True:
            asset_data = self.get()
        if asset_data is None:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"Cannot find asset {self.id}", "INVALID_REQUEST", 1))

        if self.state is not None and asset_data.state!=self.state:
            asset_data.state = self.state
        if self.adgroup is not None and asset_data.adgroup is None:
            asset_data.adgroup = self.adgroup
        if self.account is not None and asset_data.account is None:
            asset_data.account = self.account
        if self.customer is not None and asset_data.customer is None:
            asset_data.customer = self.customer
        copy = None
        if asset_data.adgroup is not None and asset_data.adgroup.clientCustomerId!=asset_data.clientCustomerId:
            # print("need to copy and create asset", asset_data.id)
            asset_copy = asset_data.to_dict().copy()
            copy = asset_data.id
            if "id" in asset_copy:
                del asset_copy["id"]
            if "assetId" in asset_copy:
                del asset_copy["assetId"]
            asset_copy["name"] = f"Asset copy {uuid.uuid1()}"
            new_asset = Asset.from_dict(asset_copy)
            new_asset.adgroup = asset_data.adgroup
            new_asset.clientCustomerId = asset_data.adgroup.clientCustomerId
            asset_data = new_asset
            # return ApiResponse(ResponseStatus.ERROR, StatusCode.status_500, None, Error(f"need to create new asset {json.dumps(new_asset.to_dict())}", "INVALID_REQUEST", 1))
        if asset_data.check_eligibility(file).status == ResponseStatus.OK and copy is None:
            return ApiResponse(ResponseStatus.OK, StatusCode.status_200, Asset.from_dict(asset_data.to_dict()), None)
        parsable = asset_data.build_parsable(file)
        if parsable.status == ResponseStatus.ERROR:
            return parsable
        _file: File = parsable.data
        # print('parsable created file', _file.id, self.id)
        asset_data.hash = _file.id
        setup = asset_data.setup_asset(_file)
        if setup.status == ResponseStatus.ERROR:
            return setup
        id = asset_data.id
        # if copy is not None:
        #     print("copyed", copy, id)
        build = None
        if bool(id) is False:
            id = asset_data.setup_asset_id(file)
            # print("new id", id)
            # print("new imageUrl", asset_data.imageUrl)
            # print("new url", asset_data.url)
            data_asset = Asset({"id": id}).get()
            # print("data_asset", data_asset)
            if data_asset is not None:
                if data_asset.check_eligibility(file).status == ResponseStatus.OK:
                    return ApiResponse(ResponseStatus.OK, StatusCode.status_302, data_asset, None)
                asset_data = data_asset

        build = asset_data.handle_asset(file=_file, test = test, cropAsset=cropAsset);
        if build.status == ResponseStatus.OK:
            asset: Asset = build.data;
            # print("asset passed build successfully", asset.id)
            if bool(asset.id) is False:
                asset.id = id
                if createAsset is True:
                    # print('creating not created asset done', asset.id)
                    create = self.create(asset.to_dict(), asset.accountId, _file, asset.clientCustomerId, False, test)
                    # print('creating not created asset done', create.data.id if create.data is not None else create.data)
                    return create
        return build
        # return ApiResponse(ResponseStatus.ERROR, StatusCode.status_500, None, Error(f"Cannot build asset {self.id}", "INVALID_REQUEST", 1))

    def handle_asset(self, file: File | None=None, cropAsset=True, test = False):
        asset = None
        eligibility = self.check_eligibility(file)
        if eligibility.status_code == StatusCode.status_200:
            if self.need_crop() is True:
                # print('need crop', self.id)
                # estimated_file_size = estimate_cropped_file_size(file.width, file.height, file.size, self.crop.x, self.crop.y, self.crop.xPosition, self.crop.yPosition)
                # print('estimated_file_size', estimated_file_size)
                self.crop.xPosition = self.crop.x + self.crop.xPosition
                self.crop.yPosition = self.crop.y + self.crop.yPosition
                if cropAsset is True:
                    crop = crop_image(
                        self.url, self.accountId, self.crop.to_dict(), self.collection_name
                    )
                    # print('crop', crop)
                    if isinstance(crop, dict) is True and bool(crop) is True:
                        self.size = crop['size']
                        new_asset = {**self.to_dict().copy(), **crop}
                        new_asset['imageUrl'] = dict(crop).get('file_url')
                        new_asset["crop_state"] = "done"
                        if bool(self.id):
                            new_asset["id"] = self.id
                        if test is True:
                            asset = Asset.from_dict(new_asset.copy())
                        if bool(self.id) is True:
                            update = self.update(new_asset, True)
                            if update.status == ResponseStatus.OK:
                                asset = Asset.from_dict(new_asset.copy())
                        else:
                            print('asset not created yet we need to skip the update')
                            asset = Asset.from_dict(new_asset.copy())

                else:
                    # print("skipping asset crop")
                    asset = Asset.from_dict(self.to_dict())
                # asset = Asset.from_dict(self.to_dict())
            else:
                # print('asset no need crop')
                if file is not None:
                    file_data = DictUtils.merge_dict(DictUtils.omit_fields(file.to_dict(), ['id', 'createdAt', 'createdBy']), self.to_dict(), True)
                    asset = Asset.from_dict(file_data)
                    asset.imageUrl = file.file_url
                    asset.size = file.size
                    asset.hash = file.id
                    asset.url = file.url
                    asset.content_type = file.content_type
                else:
                    asset = Asset.from_dict(self.to_dict())
                asset.usage = self.usage
                asset.useFor = self.useFor
                asset.owner = self.owner
                asset.accountId = self.accountId
                asset.crop_state = "done"
            return ApiResponse(ResponseStatus.OK, StatusCode.status_200, asset, None)
        else:
            return eligibility
            

    def getHash(self):
        if bool(self.hash) is True:
            file = File({"id", self.hash}).get()
            return File
        return None

    @staticmethod
    def find_customer(id: str):
        if id is None or bool(id) is False:
            return ApiResponse(
                ResponseStatus.ERROR,
                StatusCode.status_400,
                None,
                Error("Customer not found", "INVALID_REQUEST", 1),
            )

        client_customer_id = ClientCustomerId(id).get()
        if client_customer_id is None:
            return ApiResponse(
                ResponseStatus.ERROR,
                StatusCode.status_400,
                None,
                Error("Customer not found", "INVALID_REQUEST", 1),
            )
        return ApiResponse(
            ResponseStatus.OK, StatusCode.status_200, client_customer_id, None
        )
    def getAssets(self, aacid, fields=None):
        assets = []
        assets_ = self.query(Asset, "asset", query_params=[{"key": "accountId", "comp": "==", "value": aacid}])
        request_fields = filter_asset_request_fields(fields)
        # print('account query', account)
        for asset in assets_:
            assets.append(asset)
            # campaigns.append(D(campaign.to_dict(request_fields)))

        return assets

    @staticmethod
    def findAsset(id, aacid=None):
        isID = False
        try:
            id = int(id)
        except Exception as e:
            isID = True
        if isID is True:
            asset = Asset.from_dict({"id": id}).get(ASSETS_PICKABLE_FIELDS)
            if asset is None:
                return None
            return asset
        if (
            id is None
            or aacid is None
            or (isinstance(id, str) and len(id) == 0)
            or (isinstance(aacid, str) and len(aacid)) == 0
            or (isinstance(id, int) and id == 0)
        ):
            return None
        query = [
                {"key": "accountId", "comp": "==", "value": aacid},
                {"key": "assetId", "comp": "==", "value": id},
            ]
        query_key = hashlib.sha1(str(json.dumps(query)).encode()).hexdigest()
        asset_cache = cache.get(query_key)
        if asset_cache is not None:
            return Asset.from_dict(asset_cache)
        asset = Asset(None).query(
            Asset,
            "asset",
            query_params=query,
            first=True,
        )
        cache.set(query_key, asset)
        return asset

    def validate_customer(self):
        validate = Asset.find_customer(self.clientCustomerId)
        return validate

    def to_dict(self):
        fullSize = None
        if self.fullSizeInfo is not None:
            fullSize = self.fullSizeInfo.to_dict()
        crop = None
        if bool(self.crop) is True:
            crop = self.crop.to_dict()
        return DictUtils.remove_none_values(
            {
                "id": self.id,
                "name": self.name,
                "accountId": self.accountId,
                "owner": self.owner,
                "clientCustomerId": self.clientCustomerId,
                "assetId": self.assetId,
                "imageMimeType": self.imageMimeType,
                "imageFileSize": self.imageFileSize,
                "assetName": self.assetName,
                "assetSubType": self.assetSubType,
                "usage": self.usage,
                "useFor": self.useFor,
                "width": self.width,
                "height": self.height,
                "imageUrl": self.imageUrl,
                "file_url": self.file_url,
                "type": self.type,
                "size": self.size,
                "url": self.url,
                "originalImageInfo": self.originalImageInfo,
                "fullSizeInfo": fullSize,
                "crop": crop,
                "data": self.data,
                "hash": self.hash,
                "state": self.state,
                "content_type": self.content_type,
                "customer": self.customer.to_dict() if self.customer is not None else None,
                "account": self.account.to_dict() if self.account is not None else None,
                "adgroup": self.adgroup
            }
        )

    def get(self, fields=None):
        id = getattr(self, "id", None)
        assetId = getattr(self, "assetId", None)
        asset: Asset | None = None
        if bool(id) is False and bool(assetId) is True:
            query_array = [{"key": "assetId", "comp": "==", "value": assetId}]
            query_cache_key = hashlib.sha1(str(json.dumps(query_array)).encode()).hexdigest()
            asset = self.cache.get(query_cache_key)
            if asset is not None:
                asset.correct_crop()
                asset.correct_use_for()
                return asset
            asset = Asset(asset=None).query(
                Asset,
                "asset",
                query_params=[{"key": "assetId", "comp": "==", "value": assetId}],
            )
            asset.correct_crop()
            asset.correct_use_for()
            self.cache.set(query_cache_key, asset)
            return asset
        if bool(id) is None:
            return None
        asset = self.cache.get(self.id)
        if asset is not None:
            return asset
        doc = self.document_reference().get()
        if doc.exists is False:
            return None
        data = {"id": doc.id, **doc.to_dict()}
        asset = Asset.from_dict(
            asset=Asset.from_dict(asset=data).to_dict(),
            db=self.db,
            collection_name=self.collection_name,
        )
        asset.correct_crop()
        asset.correct_use_for()
        self.cache.set(asset.id, asset)
        return asset

    # def crop_image(self):
    def make_asset(self, client_customer_id=None):
        if bool(self.assetId) is True:
            if (
                self.clientCustomerId == client_customer_id
                and bool(client_customer_id) is True
            ):
                return Asset.from_dict(self.to_dict())
        if bool(self.imageUrl) is False:
            if bool(self.crop) is True:
                return None

    def need_crop(self):
        if not self.hasCrop():
            return False
        if self.crop_state is not None and self.crop_state.lower() == "done":
            return False
        return True

    def determine_dimensions(self, crop: AssetCrop):
        if crop is not None and bool(crop.to_dict()) is True:
            width = crop.xPosition
            height = crop.yPosition
            return (width, height)

    def _check_aspect_ratio(self, width: int, height: int, target_ratio: str) -> bool:
        """
        Check if image aspect ratio matches the target ratio.
        
        Args:
            target_ratio: String like "1.91:1", "1:1", "4:3", etc.
        
        Returns:
            bool: True if aspect ratio is within acceptable range
        """
        try:
            if not target_ratio or not hasattr(self, 'width') or not hasattr(self, 'height'):
                return True  # Skip ratio check if not specified
            
            # Parse target ratio
            if ':' in target_ratio:
                target_w, target_h = map(float, target_ratio.split(':'))
                target_ratio_value = target_w / target_h
            else:
                # Handle single number ratios (e.g., "1.91" means 1.91:1)
                target_ratio_value = float(target_ratio)
            
            # Calculate current image ratio
            current_ratio = width / height
            
            # Allow some tolerance (e.g., ±2%)
            tolerance = 0.02
            min_ratio = target_ratio_value * (1 - tolerance)
            max_ratio = target_ratio_value * (1 + tolerance)
            
            ratio_ok = min_ratio <= current_ratio <= max_ratio
            
            # Debug info
            # if not ratio_ok:
            #     print(f"Aspect ratio mismatch: current={current_ratio:.3f}, target={target_ratio_value:.3f}, tolerance=±{tolerance*100}%")
            
            return ratio_ok
            
        except Exception as e:
            print(f"Error checking aspect ratio: {e}")
            return True  # Skip ratio check on error
    
    def correct_crop(self):
        if not self.hasCrop() or None in [self.width, self.height]:
            return;
        x = self.crop.x
        y = self.crop.y
        xPosition = self.crop.xPosition
        yPosition = self.crop.yPosition
        xDiff = xPosition - x
        yDiff = yPosition - y
        if xDiff == self.width:
            xPosition =  xDiff
        if yDiff == self.height:
            yPosition =  yDiff
        self.crop.xPosition = xPosition
        self.crop.yPosition = yPosition
        
    def correct_use_for(self):
        width = self.width
        height = self.height
        if self.hasCrop():
            width = self.crop.xPosition
            height = self.crop.yPosition
        if self._check_aspect_ratio(width, height, "1:1"):
            self.useFor = "sqr"
        elif self._check_aspect_ratio(width, height, "4:1"):
            self.useFor = "rect"
        elif self._check_aspect_ratio(width, height, "1.91:1"):
            self.useFor = "rect"
    
    def compare_size(self, v, file: File | None = None):
        # print('compare', self.id, self.need_crop())
        # if self.id =="03e638a2020758dc9cb3c18286dc5b0a0fe90677":
        #     print('compare', self.need_crop(), self.hasCrop())
        if self.need_crop() is True:
            self.correct_crop()
            crop = self.crop.to_dict().copy()
            if bool(crop) is True:
                if self.crop.xPosition > v['maxWidth'] or self.crop.yPosition > v['maxHeight']:
                    return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"File width or height is too large for crop {self.crop.xPosition}x{self.crop.yPosition} over {v['maxWidth']}x{v['maxHeight']} for {self.id if self.id is not None else self.state if self.state is not None else self.assetId if self.assetId is not None else 'unknown'}", "INVALID_REQUEST", 1))
                if self.crop.xPosition < v['minWidth'] or self.crop.yPosition < v['minHeight']:
                    return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"File width or height is too small for crop {self.crop.xPosition}x{self.crop.yPosition} over {v['minWidth']}x{v['minHeight']} for {self.id if self.id is not None else self.state if self.state is not None else self.assetId if self.assetId is not None else 'unknown'}", "INVALID_REQUEST", 1))
                ratio_ok = self._check_aspect_ratio(self.crop.xPosition, self.crop.yPosition, v['ratio'])
                if ratio_ok is False:
                    return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"Invalid asset ({self.id}) aspect ratio ({v['ratio']}) ({json.dumps(self.crop.to_dict())}) for {self.id if self.id is not None else self.state if self.state is not None else self.assetId if self.assetId is not None else 'unknown'}", "INVALID_REQUEST", 1))
        else:
            check = [bool(self.size), bool(self.width), bool(self.height)]
            width = 0
            height = 0
            size = 0
            if False in check:
                if file is not None:
                    # print('using file', file.file_url)
                    width = file.width
                    height = file.height
                    size = file.size
            else:
                # print('using self')
                width = self.width
                height = self.height
                size = self.size
            if self.hasCrop():
                self.correct_crop()
                width = self.crop.xPosition
                height = self.crop.yPosition
            # print(width, height, size)
            if size is not None and size // 1024 > v['maxFileSize']:
                return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"File size is too large {size // 1024}Ko over {v['maxFileSize']}Ko", "INVALID_REQUEST", 1))
            if width < v['minWidth'] or height < v['minHeight']:
                return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"File width or height is too small ({width}x{height}) for {self.id if self.id is not None else self.state if self.state is not None else 'asset'} over {v['minWidth']}x{v['minHeight']}", "INVALID_REQUEST", 1))
            if width > v['maxWidth'] or height > v['maxHeight']:
                return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"File width or height is too large ({width}x{height}) max ({v['maxWidth']}x{v['maxHeight']}) for {self.id if self.id is not None else self.state if self.state is not None else 'asset'}", "INVALID_REQUEST", 1))
            ratio_ok = self._check_aspect_ratio(width, height, v['ratio'])
            if ratio_ok is False:
                return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"Invalid asset ({self.id}) aspect ratio ({v['ratio']}) ({width}x{height}) for {self.id if self.id is not None else self.state if self.state is not None else 'asset'}", "INVALID_REQUEST", 1))
        return ApiResponse(ResponseStatus.OK, StatusCode.status_200, None, None)
    def check_eligibility(self, file: File | None = None):
        # size_before = get_file_size_from_url(file.file_url)
        # print('size_before', size_before)
        is_valid = None
        error = None
        for usage in self.usage:
            # print('usage', usage)
            asset_type = "marketing_image"
            if usage == "image":
                if self.useFor == "sqr":
                    asset_type = "square_marketing_image"
            else:
                asset_type = "logo_landscape"
                if self.useFor == "sqr":
                    asset_type = "logo"
            v = pydash.find(IMAGES_RATIO, lambda x: x["type"] == asset_type)
            # print('type', asset_type)
            is_valid = self.compare_size(v, file)
            if is_valid.status_code != 200:
                error = is_valid
                break
        # file_size = estimate_pre_crop_file_size(file.width, file.height,file.size, self.crop.xPosition, self.crop.yPosition)
        # print('estimated file_size', file_size)
        if error is not None:
            return error        
        return ApiResponse(ResponseStatus.OK, StatusCode.status_200, None, None)

    def checkpoint_asset_account(self, act: str | Account | None = None, state: str | None = None):
        error = {"state": state if state is not None else self.state}
        account = self.account;
        id = self.accountId
        if  bool(id) is False:
            if act is not None:
                if isinstance(act, str):
                    id = act
                else:
                    id = act.id
        if bool(id) and (account is None or bool(account) is False or bool(account.id) is False):
            account = Account(id).get()
        if account is not None:
            self.account = account
            return ApiResponse(ResponseStatus.OK, StatusCode.status_200, account, None)
        error["message"] = f"Cannot found account for asset {self.id}"
        return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(error, "INVALID_REQUEST", 1))
    def checkpoint_asset_customer(self, customer: str | ClientCustomerId | None = None, state: str | None = None):
        error = {"state": self.state if self.state is not None else state}
        clientCustomer = self.customer
        if clientCustomer is not None:
            return ApiResponse(ResponseStatus.OK, StatusCode.status_200, clientCustomer, None)
        id = self.clientCustomerId
        if id is None:
            if customer is not None:
                if isinstance(customer, str):
                    id = customer
                else:
                    id = customer.id
        if clientCustomer is None or bool(clientCustomer) is False or bool(clientCustomer.id) is False:
            if bool(id):
                clientCustomer = ClientCustomerId(id).get()
            else:
                customers = ClientCustomerId().getDefault()
                if bool(customers) is True:
                    clientCustomer = customers[0]
                    self.customer = clientCustomer
                    return ApiResponse(ResponseStatus.OK, StatusCode.status_200, clientCustomer, None)
        if clientCustomer is not None:
            self.customer = clientCustomer
            return ApiResponse(ResponseStatus.OK, StatusCode.status_200, clientCustomer, None)
        error["message"] = f"Cannot found customer {id} for asset {self.id}"
        return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(error, "INVALID_REQUEST", 1))

    def setup_asset_id(self, file: File | None = None):
        id_data = DictUtils.pick_fields(self.to_dict(), ["imageUrl", "url", "accountId", "owner"])
        if bool(self.assetId) is True:
            id_data["assetId"] = self.assetId
        if self.hasCrop():
            id_data["crop"] = self.crop.to_dict()
        else:
            if file is not None:
                id_data["width"] = file.width
                id_data["height"] = file.height
        return hashlib.sha1(str(json.dumps(id_data)).encode()).hexdigest()

    def setup_asset(self, file: File | None):
        account_check = self.checkpoint_asset_account()
        if account_check.status == ResponseStatus.ERROR:
            return account_check
        account: Account = account_check.data
        customer_check = self.checkpoint_asset_customer()
        if customer_check.status == ResponseStatus.ERROR:
            return customer_check
        customer: ClientCustomerId = customer_check.data
        self.accountId = account.id
        self.clientCustomerId = customer.id
        self.owner = account.owner
        self.createdBy = f"users/{account.owner}"
        if self.need_crop():
            dim = self.determine_dimensions(self.crop)
            self.width = dim[0]
            self.height = dim[1]
        else:
            if file is not None:
                self.width = file.width
                self.height = file.height
        return ApiResponse(ResponseStatus.OK, StatusCode.status_200, Asset.from_dict(self.to_dict()), None)


    
    
    def hasCrop(self):
        if bool(self.crop) is False:
            return False
        return self.crop.x is not None and self.crop.y is not None and self.crop.xPosition is not None and self.crop.yPosition is not None

        
    def prepare_asset(self, data, _account: Account | None, file: File | None=None, _customerId=None, testMode=False):
        state = dict(data).get("state", None)
        error = {"state": state}
        data_create = DictUtils.dict_from_keys(
            data, AssetsFields().filtered_keys("mutable", True)
        )
        if bool(data_create) is False:
            error["message"] = "Invalid request required valid mutable fields"
            return ApiResponse(
                ResponseStatus.ERROR,
                StatusCode.status_400,
                None,
                Error(error, "INVALID_REQUEST", 1),
            )
        asset_create = Asset.from_dict(data_create.copy())
        asset_create.state = state
        data_asset = asset_create.build_asset(file, True, False, False, testMode)
        if data_asset.status_code != 200:
            return data_asset
        print('second eligibility', data_asset.data.to_dict())
        validity = data_asset.data.check_eligibility(file)
        if validity.status_code != 200:
            print('validity', validity.error.message)
            error["message"] = validity.error.message
            return ApiResponse(
                ResponseStatus.ERROR,
                StatusCode.status_400,
                None,
                Error(error, "INVALID_ASSET", 1),
            )
        # data_create = data_asset.to_dict()

        # document_id = self.collection().document().id
        # if "assetId" in data and bool(data["assetId"]) is True:
        #     document_id = str(data["assetId"])
        model = Asset.generate_model()
        create = DictUtils.merge_dict(data_asset.data.to_dict(), model, True)
        asset = Asset.from_dict(create)
        if state is not None and bool(state) is True:
            asset.state = str(state)
        # print('asset valid', asset.data)
        return ApiResponse(
            ResponseStatus.OK, StatusCode.status_200, asset, None
        )
    def create(self, data: dict, _account, file: File=None, _customerId=None, prepare = True, testMode=False):
        # creation_date = created_at * 1000
        # return DateUtils.from_timestamp(creation_date).isoformat()
        # try:
        # print('campaign account', data)
        values = None
        state = dict(data).get("state", None)
        error = {"state": state}
        if bool(prepare) is True:
            prepared = self.prepare_asset(data, _account, file, _customerId, testMode)
            if prepared.status_code == StatusCode.status_200:
                values = prepared.data
            else:
                if prepared.status_code == StatusCode.status_302:
                    prepared.status_code = 200
                return prepared
        else:
            values = Asset.from_dict(data)
        document_id = None
        if bool(values.id) is False:
            error["message"] = "Invalid request id not present"
            return ApiResponse(
                ResponseStatus.ERROR,
                StatusCode.status_400,
                None,
                Error(error, "INVALID_REQUEST", 1),
            )
        document_id = str(values.id)

        model = Asset.generate_model()
        data_create = DictUtils.merge_dict(values.to_dict(), model, True)
        create = {**data_create, "createdAt": firestore.SERVER_TIMESTAMP}
        del create["state"]
        if testMode is False:
            doc_ref = self.collection().document(document_id)
            doc_ref.set(create, True)
        # print("create", create)
        asset = Asset.from_dict(create)
        print("create asset successfully done", asset.id)
        if state is not None and bool(state) is True:
            asset.state = str(state)
        asset.correct_use_for()
        self.cache.set(asset.id, asset)
        return ApiResponse(
            ResponseStatus.OK, StatusCode.status_200, asset, None
        )
        # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the created campaign object.  # Return the
        # except Exception as e:
        #     print(e)
        #     return {"error": "Exception creating campaign"};

    def update(self, data, silent=True):
        try:
            data_update = {}
            changed_fields = []
            if silent is True:
                data_update = pydash.pick(
                    data, AssetsFields().filtered_keys("pickable", True)
                )
            else:
                last_value = self.to_dict()
                print("data update", data)
                # print('last', last_value);
                filtered_value = pydash.pick(
                    data, AssetsFields().filtered_keys("editable", True)
                )
                # print('filtered', filtered_value);
                new_value = DictUtils.merge_dict(filtered_value, last_value, True)
                _changed_fields = DictUtils.compare_nested_objects(
                    last_value, new_value
                )
                changed_fields = ArrayUtils.filter(
                    DictUtils.get_keys(_changed_fields),
                    lambda x: str(x).find(".") == -1,
                )
                data_update = DictUtils.dict_from_keys(filtered_value, changed_fields)
            if bool(data_update) is False:
                return ApiResponse(
                    ResponseStatus.OK, StatusCode.status_200, data_update, None
                )
            # try:
            #     self.document_reference().set(data_update, merge=True)
            # except Exception as e:
            #     return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(str(e),"INVALID_REQUEST", 1))
            # return DictUtils.dict_from_keys(self.get(), changed_fields);
            # fields = changed_fields
            print("data_update asset", data_update)
            return ApiResponse(
                ResponseStatus.OK,
                StatusCode.status_200,
                DictUtils.pick_fields(data_update, ASSETS_PICKABLE_FIELDS),
                None,
            )
        except Exception as e:
            print(e)
            return None

    def to_json(self, _fields=None):
        return self.to_dict()



import os
from dataclasses import dataclass

from ......utils.utils import BaseClass, BaseFieldsClass, DictUtils

ASSETS_COLLECTION = os.environ.get('GADS_ASSETS_COLLECTION');


@dataclass
class AssetCropFields(BaseFieldsClass):
    x = "x"
    y = "y"
    xPosition = "xPosition"
    yPosition = "yPosition"
    crop_based_width = "crop_based_width"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.class_fields_props = AssetCropFieldsProps
@dataclass
class AssetFullSizeInfoFields(BaseFieldsClass):
    imageWidth = "imageWidth"
    imageHeight = "imageHeight"
    imageUrl = "imageUrl"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.class_fields_props = AssetsFieldsProps
@dataclass
class AssetsFields(BaseFieldsClass):
    id = "id"
    accountId = "accountId"
    name = "name"
    assetId = "assetId"
    owner = "owner"
    clientCustomerId = "clientCustomerId"
    imageMimeType = "imageMimeType"
    imageFileSize = "imageFileSize"
    assetName = "assetName"
    assetSubType = "assetSubType"
    fullSizeInfo = "fullSizeInfo"
    size = "size"
    file_url = "file_url"
    type = "type"
    data = "data"
    owner = "owner"
    usage = "usage"
    useFor = "useFor"
    width = "width"
    height = "height"
    imageUrl = "imageUrl"
    originalImageInfo = "originalImageInfo"
    crop = "crop"
    crop_state = "crop_state"
    content_type = "content_type"
    url = "url"
    hash = "hash"
    state = "state"
    account = "account"
    customer = "customer"
    adgroup = "adgroup"
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.class_fields_props = AssetsFieldsProps

AssetCropFieldsProps = {
    AssetCropFields.x: {
        "internal": True,
        "type": int,
        "required": True,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    AssetCropFields.y: {
        "internal": True,
        "type": int,
        "required": True,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    AssetCropFields.xPosition: {
        "internal": True,
        "type": int,
        "required": True,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    AssetCropFields.yPosition: {
        "internal": True,
        "type": int,
        "required": True,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    AssetCropFields.crop_based_width: {
        "internal": True,
        "type": int,
        "required": False,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
}
AssetFullSizeInfoFieldsProps = {
    AssetFullSizeInfoFields.imageWidth: {
        "internal": True,
        "type": int,
        "required": True,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    AssetFullSizeInfoFields.imageHeight: {
        "internal": True,
        "type": int,
        "required": True,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    AssetFullSizeInfoFields.imageUrl: {
        "internal": True,
        "type": str,
        "required": True,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    }
}

AssetsFieldsProps = {
    AssetsFields.id: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    AssetsFields.hash: {
        "internal": True,
        "type": str,
        "required": True,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    AssetsFields.size: {
        "internal": True,
        "type": int,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    AssetsFields.accountId: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    AssetsFields.name: {
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    AssetsFields.assetId: {
        "type": int,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    AssetsFields.owner: {
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    AssetsFields.clientCustomerId: {
        "type": str,
        "required": False,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    AssetsFields.imageMimeType: {
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    AssetsFields.usage: {
        "type": list,
        "required": False,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    AssetsFields.useFor: {
        "type": str,
        "required": False,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    AssetsFields.width: {
        "type": [int, str],
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    AssetsFields.height: {
        "type": [int, str],
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    AssetsFields.imageUrl: {
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    AssetsFields.originalImageInfo: {
        "type": str,
        "required": False,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    AssetsFields.assetName: {
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    AssetsFields.assetSubType: {
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    AssetsFields.imageFileSize: {
        "type": [int, str],
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    AssetsFields.fullSizeInfo: {
        "type": dict,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    AssetsFields.crop: {
        "type": dict,
        "required": False,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    AssetsFields.data: {
        "type": str,
        "required": False,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    AssetsFields.crop_state: {
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    AssetsFields.content_type: {
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    AssetsFields.url: {
        "type": str,
        "required": False,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    AssetsFields.state: {
        "type": str,
        "required": False,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    },
    AssetsFields.account: {
        "type": dict,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": False,
        "default_value": None
    },
    AssetsFields.customer: {
        "type": dict,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": False,
        "default_value": None
    },
    AssetsFields.adgroup: {
        "type": dict,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": False,
        "default_value": None
    },

}

STANDARD_FIELDS = AssetsFields(fields_props=AssetsFieldsProps).filtered_keys('pickable', True)
ASSET_CROP_FIELDS = AssetCropFields(fields_props=AssetCropFieldsProps).filtered_keys('pickable', True)
FULL_SIZE_INFO_FIELDS = AssetFullSizeInfoFields(fields_props=AssetFullSizeInfoFieldsProps).filtered_keys('pickable', True)
ASSETS_PICKABLE_FIELDS = AssetsFields(fields_props=AssetsFieldsProps).filtered_keys('pickable', True)
ASSETS_NOT_PICKABLE_FIELDS = AssetsFields(fields_props=AssetsFieldsProps).filtered_keys('pickable', False)



from typing import Any, Dict, List, Union, Set
import re

def _resolve_path(path: str) -> List[Union[str, int]]:
    """Convert string path to list of keys/indices."""
    if not path:
        return []
    
    parts = []
    # Handle both array[0] and simple dot notation
    matches = re.findall(r'([^.\[\]]+)|\[(\d+)\]', path)
    
    for match in matches:
        # Each match is a tuple of two groups
        if match[0]:  # Normal key
            parts.append(match[0])
        else:  # Array index
            parts.append(int(match[1]))
            
    return parts

def _get_value(obj: Any, path: List[Union[str, int]], collect_arrays: bool = True) -> Any:
    """
    Get value from object using path list.
    If collect_arrays is True, it will collect values from all array elements when encountering an array.
    """
    current = obj
    
    for i, key in enumerate(path):
        try:
            if isinstance(current, list) and collect_arrays and not isinstance(key, int):
                # Map operation over all array elements
                return [_get_value(item, path[i:], collect_arrays) for item in current]
            elif isinstance(current, (dict, list)):
                current = current[key]
            elif hasattr(current, str(key)):
                current = getattr(current, str(key))
            else:
                return None
        except (KeyError, IndexError, TypeError, AttributeError):
            return None
            
    return current

def _set_value(target: dict, path: List[Union[str, int]], value: Any) -> None:
    """Set value in target dict using path list."""
    if not path:
        return
        
    current = target
    
    for i, key in enumerate(path[:-1]):
        if isinstance(key, int):
            prev_key = path[i-1] if i > 0 else None
            if prev_key is not None:
                if prev_key not in current or not isinstance(current[prev_key], list):
                    current[prev_key] = []
                while len(current[prev_key]) <= key:
                    current[prev_key].append({})
                current = current[prev_key][key]
            else:
                if not isinstance(current, list):
                    current = target.setdefault(key, {})
                else:
                    while len(current) <= key:
                        current.append({})
                    current = current[key]
        else:
            current = current.setdefault(key, {})

    last_key = path[-1]
    if isinstance(value, list) and all(isinstance(v, dict) for v in value):
        # Handle array results from collection
        if last_key not in current:
            current[last_key] = value
        else:
            if not isinstance(current[last_key], list):
                current[last_key] = value
            else:
                # Merge with existing array
                while len(current[last_key]) < len(value):
                    current[last_key].append({})
                for i, v in enumerate(value):
                    current[last_key][i].update(v)
    else:
        current[last_key] = value

def pick(obj: Any, *paths: Union[str, List[str], Set[str], tuple]) -> dict:
    """
    Creates an object composed of the picked object properties.
    Supports deep picking using dot notation and array indices.
    When encountering arrays, it will collect the specified property from all elements.
    
    Args:
        obj: Source object
        *paths: String paths or list/set/tuple of paths
        
    Returns:
        dict: New object with picked properties
        
    Examples:
        >>> data = {'users': [{'name': 'John'}, {'name': 'Jane'}]}
        >>> pick(data, 'users.name')
        {'users': [{'name': 'John'}, {'name': 'Jane'}]}
    """
    if not obj:
        return {}
    
    # Flatten paths if a list/set/tuple was passed
    flat_paths = []
    for path in paths:
        if isinstance(path, (list, set, tuple)):
            flat_paths.extend(path)
        else:
            flat_paths.append(path)
    
    result = {}
    for path in flat_paths:
        resolved_path = _resolve_path(path)
        value = _get_value(obj, resolved_path)
        if value is not None:
            _set_value(result, resolved_path, value)
            
    return result

# Example usage
if __name__ == "__main__":
    data = {
        'users': [
            {
                'id': 1,
                'name': 'John',
                'contacts': [
                    {'type': 'email', 'value': 'john@example.com'},
                    {'type': 'phone', 'value': '123-456-7890'}
                ]
            },
            {
                'id': 2,
                'name': 'Jane',
                'contacts': [
                    {'type': 'email', 'value': 'jane@example.com'},
                    {'type': 'phone', 'value': '098-765-4321'}
                ]
            }
        ],
        'settings': {
            'notifications': True
        }
    }

    # Test cases
    print("Pick all names from users:")
    print(pick(data, 'users.name'))
    # {'users': [{'name': 'John'}, {'name': 'Jane'}]}

    # print("\nPick all contact values:")
    # print(pick(data, 'users.contacts.value'))
    # # {'users': [{'contacts': [{'value': 'john@example.com'}, {'value': '123-456-7890'}]}, 
    # #            {'contacts': [{'value': 'jane@example.com'}, {'value': '098-765-4321'}]}]}

    # print("\nPick multiple properties:")
    # print(pick(data, 'users.name', 'users.id'))
    # # {'users': [{'name': 'John', 'id': 1}, {'name': 'Jane', 'id': 2}]}

    # print("\nMix of array and non-array paths:")
    # print(pick(data, 'users.name', 'settings.notifications'))
    # # {'users': [{'name': 'John'}, {'name': 'Jane'}], 'settings': {'notifications': True}}

    # # Additional test case with nested arrays
    # data2 = {
    #     'departments': [
    #         {
    #             'name': 'Engineering',
    #             'teams': [
    #                 {'name': 'Frontend', 'members': ['Alice', 'Bob']},
    #                 {'name': 'Backend', 'members': ['Charlie', 'David']}
    #             ]
    #         },
    #         {
    #             'name': 'Marketing',
    #             'teams': [
    #                 {'name': 'Digital', 'members': ['Eve', 'Frank']},
    #                 {'name': 'Brand', 'members': ['Grace', 'Henry']}
    #             ]
    #         }
    #     ]
    # }

    # print("\nPick nested array properties:")
    # print(pick(data2, 'departments.teams.name'))
    # # {'departments': [
    # #     {'teams': [{'name': 'Frontend'}, {'name': 'Backend'}]},
    # #     {'teams': [{'name': 'Digital'}, {'name': 'Brand'}]}
    # # ]}


from .models import (
    WorkflowWorkspace,
    WorkflowWorkspaceFields,
    WORKSPACE_PICKABLE_FIELDS,
    WORKSPACE_STATUS,
    WorkflowScenario,
    WorkflowScenarioFields,
    SCENARIO_PICKABLE_FIELDS,
    WorkflowCreative,
    WorkflowCreativeFields,
    CREATIVE_PICKABLE_FIELDS,
    WorkflowDiscussionMessage,
    DISCUSSION_PICKABLE_FIELDS,
    WorkflowChannel,
    WorkflowChannelFields,
    CHANNEL_PICKABLE_FIELDS,
    CHANNEL_TYPES,
    WorkflowChannelMessage,
    MESSAGE_PICKABLE_FIELDS,
    WorkflowDocument,
    WorkflowDocumentFields,
    DOCUMENT_PICKABLE_FIELDS,
    WorkflowMember,
    WorkflowMemberFields,
    MEMBER_PICKABLE_FIELDS,
    MEMBER_ROLES,
    WorkflowTeam,
    WorkflowTeamFields,
    TEAM_PICKABLE_FIELDS,
    WorkflowInvite,
    INVITE_PICKABLE_FIELDS,
    INVITE_STATUSES,
    WorkflowCampaignLink,
    CAMPAIGN_LINK_PICKABLE_FIELDS,
)



import os
import json
import time
from datetime import datetime, timedelta

from authlib.jose import jwt

# Re-use key loading from grant module
from ..models.grant import read_private_key, import_key


class JWKSManager:
    def get_signing_key(self):
        """Load the signing key. Priority: LOCAL_OIDC_JSON_URL → LOCAL_OIDC_PEM_URL → OIDC_JSON_URL"""
        local_json_file = os.environ.get('LOCAL_OIDC_JSON_URL', None)
        local_pem_file = os.environ.get('LOCAL_OIDC_PEM_URL', None)
        oidc_json_url = os.environ.get('OIDC_JSON_URL', None)

        if local_json_file is not None:
            try:
                fp = os.getcwd() + "/" + local_json_file
                key = import_key(fp, True)
                if key is not None:
                    return key
            except Exception as e:
                print(f'JWKSManager: failed to load local JSON key: {e}')

        if local_pem_file is not None:
            try:
                fp = os.getcwd() + "/" + local_pem_file
                key = import_key(fp, False)
                if key is not None:
                    return key
            except Exception as e:
                print(f'JWKSManager: failed to load local PEM key: {e}')

        if oidc_json_url is not None:
            import requests
            try:
                resp = requests.get(oidc_json_url)
                if resp.status_code == 200:
                    return resp.json()
            except Exception as e:
                print(f'JWKSManager: failed to fetch OIDC JSON URL: {e}')

        return read_private_key()

    def get_public_jwks(self) -> dict:
        """Return the public portion of the JWKS."""
        key = self.get_signing_key()
        if key is None:
            return {"keys": []}
        try:
            from joserfc.jwk import RSAKey
            if isinstance(key, dict) and 'kty' in key:
                rsa = RSAKey.import_key(key)
                public = rsa.as_dict(private=False)
                return {"keys": [public]}
            elif isinstance(key, dict) and 'keys' in key:
                public_keys = []
                for k in key['keys']:
                    try:
                        rsa = RSAKey.import_key(k)
                        public_keys.append(rsa.as_dict(private=False))
                    except Exception:
                        pass
                return {"keys": public_keys}
        except Exception as e:
            print(f'JWKSManager: failed to extract public key: {e}')
        return {"keys": []}

    def generate_id_token(self, client, user, scope: str, nonce: str = None,
                          auth_time: int = None, expires_in: int = 3600) -> str:
        """Generate an OIDC ID token (ES256/RS256 JWT)."""
        key = self.get_signing_key()
        if key is None:
            raise ValueError('No signing key available for ID token generation')
        now = int(time.time())
        issuer = os.environ.get('OAUTH2_ISSUER', os.environ.get('JWT_ISSUER', ''))
        payload = {
            'iss': issuer,
            'sub': str(getattr(user, 'uid', getattr(user, 'id', ''))),
            'aud': client.client_id if hasattr(client, 'client_id') else str(client),
            'iat': now,
            'exp': now + expires_in,
        }
        if nonce:
            payload['nonce'] = nonce
        if auth_time:
            payload['auth_time'] = auth_time
        # Include basic user claims when scope includes openid/profile/email
        if hasattr(user, 'email') and 'email' in scope:
            payload['email'] = user.email
        if hasattr(user, 'displayName') and 'profile' in scope:
            payload['name'] = user.displayName
        alg = os.environ.get('OIDC_JWT_ALG', 'RS256')
        header = {'alg': alg}
        token = jwt.encode(header, payload, key, check=False)
        if isinstance(token, bytes):
            return token.decode('utf-8')
        return token

    def openid_configuration(self, issuer: str = None) -> dict:
        """Return the OpenID Connect discovery document."""
        if issuer is None:
            issuer = os.environ.get('OAUTH2_ISSUER', os.environ.get('JWT_ISSUER', ''))
        return {
            "issuer": issuer,
            "authorization_endpoint": issuer + "/o/v1/authorize",
            "token_endpoint": issuer + "/o/v1/token",
            "userinfo_endpoint": issuer + "/o/v1/userinfo",
            "jwks_uri": issuer + "/.well-known/jwks.json",
            "revocation_endpoint": issuer + "/o/v1/revoke",
            "introspection_endpoint": issuer + "/o/v1/introspect",
            "device_authorization_endpoint": issuer + "/o/v1/device",
            "response_types_supported": ["code", "token", "id_token", "code token", "code id_token", "id_token token"],
            "grant_types_supported": [
                "authorization_code", "implicit", "client_credentials",
                "refresh_token", "urn:ietf:params:oauth:grant-type:device_code"
            ],
            "subject_types_supported": ["public"],
            "id_token_signing_alg_values_supported": ["RS256", "ES256"],
            "scopes_supported": ["openid", "profile", "email"],
            "token_endpoint_auth_methods_supported": ["client_secret_basic", "client_secret_post", "none"],
            "claims_supported": ["sub", "iss", "aud", "exp", "iat", "email", "name", "nonce"],
            "code_challenge_methods_supported": ["S256"],
        }


jwks_manager = JWKSManager()



import uuid
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Any, Optional

from adafri.v1.base.firebase_collection import FirebaseCollectionBase

CREATIVE_COLLECTION = 'workflow_creatives'

CREATIVE_PICKABLE_FIELDS = [
    'id', 'workspaceId', 'scenarioId', 'accountId', 'campaignId',
    'type', 'name', 'headline', 'description', 'ctaText',
    'assetUrl', 'previewUrl', 'status', 'feedback',
    'createdAt', 'updatedAt', 'createdBy',
]

CREATIVE_MUTABLE_FIELDS = [
    'name', 'headline', 'description', 'ctaText',
    'assetUrl', 'previewUrl', 'status', 'feedback', 'scenarioId',
]

CREATIVE_TYPES = ['image', 'video', 'copy', 'html']
CREATIVE_STATUS = ['draft', 'in_review', 'approved', 'rejected']


def _now() -> str:
    return datetime.now(timezone.utc).isoformat()


def _new_id() -> str:
    return str(uuid.uuid4())


class WorkflowCreativeFields:
    @staticmethod
    def check_create_fields(data: dict) -> dict:
        if not data:
            return {'error': 'Missing data'}
        if not data.get('name', '').strip():
            return {'error': 'name is required'}
        if not data.get('workspaceId', '').strip():
            return {'error': 'workspaceId is required'}
        if data.get('type') not in CREATIVE_TYPES:
            return {'error': f"type must be one of: {', '.join(CREATIVE_TYPES)}"}
        return data

    @staticmethod
    def mutable_keys() -> list:
        return CREATIVE_MUTABLE_FIELDS


@dataclass(init=False)
class WorkflowCreative(FirebaseCollectionBase):
    id: str
    workspaceId: str
    scenarioId: Optional[str]
    accountId: str
    campaignId: str
    type: str
    name: str
    headline: str
    description: str
    ctaText: str
    assetUrl: str
    previewUrl: str
    status: str
    feedback: Optional[str]
    createdAt: str
    updatedAt: str
    createdBy: str

    def __init__(self, data=None, **kwargs):
        super().__init__(collection_name=CREATIVE_COLLECTION, **kwargs)
        d = data or {}
        self.id = d.get('id') or _new_id()
        self.workspaceId = d.get('workspaceId', '')
        self.scenarioId = d.get('scenarioId')
        self.accountId = d.get('accountId', '')
        self.campaignId = d.get('campaignId', '')
        self.type = d.get('type', 'image')
        self.name = d.get('name', '')
        self.headline = d.get('headline', '')
        self.description = d.get('description', '')
        self.ctaText = d.get('ctaText', '')
        self.assetUrl = d.get('assetUrl', '')
        self.previewUrl = d.get('previewUrl', '')
        self.status = d.get('status', 'draft')
        self.feedback = d.get('feedback')
        self.createdAt = d.get('createdAt') or _now()
        self.updatedAt = d.get('updatedAt') or _now()
        self.createdBy = d.get('createdBy', '')

    @classmethod
    def from_dict(cls, obj: Any, **kwargs) -> 'WorkflowCreative':
        return cls(obj, **kwargs)

    def to_dict(self) -> dict:
        return {
            'id': self.id,
            'workspaceId': self.workspaceId,
            'scenarioId': self.scenarioId,
            'accountId': self.accountId,
            'campaignId': self.campaignId,
            'type': self.type,
            'name': self.name,
            'headline': self.headline,
            'description': self.description,
            'ctaText': self.ctaText,
            'assetUrl': self.assetUrl,
            'previewUrl': self.previewUrl,
            'status': self.status,
            'feedback': self.feedback,
            'createdAt': self.createdAt,
            'updatedAt': self.updatedAt,
            'createdBy': self.createdBy,
        }

    @staticmethod
    def toListDict(items: list, fields=None) -> list:
        return [item.to_dict() for item in items]

    @classmethod
    def find(cls, creative_id: str, account_id: str = None) -> 'Optional[WorkflowCreative]':
        inst = cls()
        doc = inst.collection().document(creative_id).get()
        if not doc.exists:
            return None
        data = {**doc.to_dict(), 'id': doc.id}
        if account_id and data.get('accountId') != account_id:
            return None
        return cls(data)

    @classmethod
    def listByWorkspace(cls, workspace_id: str) -> 'list[WorkflowCreative]':
        inst = cls()
        docs = inst.collection().where('workspaceId', '==', workspace_id).stream()
        return [cls({**d.to_dict(), 'id': d.id}) for d in docs]

    @classmethod
    def listByScenario(cls, scenario_id: str) -> 'list[WorkflowCreative]':
        inst = cls()
        docs = inst.collection().where('scenarioId', '==', scenario_id).stream()
        return [cls({**d.to_dict(), 'id': d.id}) for d in docs]

    def save(self) -> 'WorkflowCreative':
        self.collection().document(self.id).set(self.to_dict())
        return self

    def update(self, data: dict) -> 'WorkflowCreative':
        filtered = {k: v for k, v in data.items() if k in CREATIVE_MUTABLE_FIELDS}
        filtered['updatedAt'] = _now()
        self.collection().document(self.id).update(filtered)
        for k, v in filtered.items():
            setattr(self, k, v)
        return self

    def delete(self) -> None:
        self.collection().document(self.id).delete()



import hashlib
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Any

from .....utils import DictUtils, Crypto, get_object_model_class, init_class_kwargs
from ....base.firebase_collection import FirebaseCollectionBase, getTimestamp
from .mfa_fields import MFAFields, MFAFieldsProps, STANDARD_FIELDS, MFA_COLLECTION
from .....utils.response import ApiResponse, Error, ResponseStatus, StatusCode


@dataclass(init=False)
class MFACredential(FirebaseCollectionBase):
    id: str
    user_uid: str
    method: str
    totp_secret: str
    phone: str
    backup_codes: list
    is_enabled: bool
    enrolled_at: str
    last_used_at: str

    def __init__(self, mfa=None, **kwargs):
        if type(mfa) is str:
            mfa = {"user_uid": mfa}
        (cls_object, keys, data_args) = init_class_kwargs(self, mfa, STANDARD_FIELDS, MFAFieldsProps, MFA_COLLECTION, ['id'], **kwargs)
        super().__init__(**data_args)
        for key in keys:
            setattr(self, key, cls_object[key])

    @staticmethod
    def generate_model(_key_="default_value"):
        model = {}
        props = MFAFieldsProps
        for k in DictUtils.get_keys(props):
            model[k] = props[k][_key_]
        return model

    @staticmethod
    def from_dict(mfa: Any = None, db=None, collection_name=None) -> 'MFACredential':
        cls_object, keys = get_object_model_class(mfa, MFACredential, MFAFieldsProps)
        return MFACredential(cls_object, db=db, collection_name=collection_name)

    def query(self, query_params: list, first=False, limit=None):
        try:
            result = []
            query_result = self.custom_query(query_params, first=first, limit=limit)
            if bool(query_result):
                if first:
                    return MFACredential.from_dict(mfa=query_result, db=self.db, collection_name=self.collection_name)
                else:
                    for doc in query_result:
                        result.append(MFACredential.from_dict(mfa=doc, db=self.db, collection_name=self.collection_name))
                    return result
            return None if first else []
        except Exception as e:
            print('Exception while querying MFACredential', e)
            return None if first else []

    def to_dict(self):
        return {
            "id": self.id,
            "user_uid": self.user_uid,
            "method": self.method,
            "totp_secret": self.totp_secret,
            "phone": self.phone,
            "backup_codes": self.backup_codes,
            "is_enabled": self.is_enabled,
            "enrolled_at": self.enrolled_at,
            "last_used_at": self.last_used_at,
        }

    def to_json(self, _fields=None):
        return self.to_dict()

    def get_by_user(self, uid: str) -> 'MFACredential':
        return self.query([{"key": MFAFields.user_uid, "comp": "==", "value": uid}], first=True)

    def create(self, **kwargs):
        model = DictUtils.merge_dict(kwargs, self.generate_model())
        obj = MFACredential.from_dict(model)
        obj.id = Crypto().generate_id(obj.user_uid + "~mfa")
        obj.enrolled_at = datetime.now(timezone.utc).isoformat()
        doc_ref = MFACredential(obj.to_json()).document_reference()
        doc_ref.set({**obj.to_json(), "createdAt": getTimestamp()}, merge=True)
        return obj

    def update(self, data: dict):
        try:
            doc_ref = self.document_reference()
            doc_ref.set(data, merge=True)
            return data
        except Exception as e:
            print('Exception while updating MFACredential', e)
            return None

    def remove(self):
        try:
            self.document_reference().delete()
            return ApiResponse(ResponseStatus.OK, StatusCode.status_200, {"message": f"MFA {self.id} deleted"}, None)
        except Exception as e:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(str(e), "INVALID_REQUEST", 1))



from ...base.firebase_collection import FirebaseCollectionBase
from ....utils import (DictUtils, DateUtils, get_object_model_class, init_class_kwargs, BaseClass, get_request_fields, RequestFields)
from ....utils.response import ApiResponse, Error, ResponseStatus, StatusCode
from .base_campaign_fields import BaseCampaignFieldsProps, STANDARD_FIELDS, BASE_CAMPAIGN_PICKABLE_FIELDS
from typing import Any
from dataclasses import dataclass

from .objective import Objective
from .adchannel import AdChannel
from .ages import Ages
from .genders import Genders
from .locations import Location
from .adschedules import AdsSchedules
from .custom_parameters import CustomParameters
from .devices import Devices


def filter_request_fields(fields, default_fields = BASE_CAMPAIGN_PICKABLE_FIELDS):
    request_fields = get_request_fields(fields, default_fields, [default_fields[0]])
    if request_fields is None or bool(request_fields) is False:
        request_fields = [default_fields[0]]
    return request_fields





@dataclass(init=False)
class BaseCampaign(FirebaseCollectionBase):
    id: str
    id_campagne: int
    name: str
    status: str
    objective: Objective
    adChannel: AdChannel
    startDate: str
    endDate: str
    startDateFormattedGoogle: str
    endDateFormattedGoogle: str
    areaTargetedOption: str
    areaExcludedOption: str
    budget: int
    budgetId: int
    dailyBudget: int
    bid: float
    strategie: str
    urlPromote: str
    ages: list[Ages]
    genders: list[Genders]
    targetedLocations: list[Location]
    excludedLocations: list[Location]
    devicesTargeted: list[Devices]
    devicesExcluded: list[Devices]
    deliveryMethod: str
    trackingTemplate: str
    finalUrlSuffix: str
    accountId: str
    type: str
    is_removed: bool
    isEnded: bool
    isComplete: bool
    isPayed: bool
    publish: bool
    publishing: bool
    budgetEnded: bool
    ad_group_id: int
    ad_group_id_firebase: str
    servingStatus: str
    owner: str
    provider: str
    urlCustomParameters: list[CustomParameters]
    adsSchedules: list[AdsSchedules]
    createdAt: any
    createdBy: any
    clientCustomerId: str
    __realBudget: int
    google_ads_campaign: dict
    _locations_targeted: bool
    _locations_excluded: bool
    _schedules_targeted: bool
    _upgraded_urls_targeted: bool
    _ads_cloned: bool = False
    api_version: str
    __baseFields = BaseCampaignFieldsProps


    def calculate_budget(self, value):
        budget = self.to_budget(value);
        return round((budget * 30)/100)
    def __init__(self, campaign=None, collection_name=None, **kwargs):
        _campaign = campaign;
        if type(campaign) is str:
            _campaign = {'id': campaign};
        (cls_object, keys, data_args) = init_class_kwargs(self, _campaign, STANDARD_FIELDS, BaseCampaignFieldsProps, collection_name, ['id'], **kwargs)
        super().__init__(**data_args);
        for key in keys:
            if key == "objective":
                setattr(self, key, Objective().from_dict(cls_object[key]))
            elif key == "adChannel":
                setattr(self, key, AdChannel().from_dict(cls_object[key]))
            elif key == "ages":
                setattr(self, key, Ages().fromListDict(cls_object[key]))
            elif key == "genders":
                setattr(self, key, Genders().fromListDict(cls_object[key]))
            elif key == "targetedLocations":
                setattr(self, key, Location().fromListDict(cls_object[key]))
            elif key == "excludedLocations":
                setattr(self, key, Location().fromListDict(cls_object[key]))
            elif key == "devicesTargeted":
                setattr(self, key, Devices().fromListDict(cls_object[key]))
            elif key == "devicesExcluded":
                setattr(self, key, Devices().fromListDict(cls_object[key]))
            elif key == "urlCustomParameters":
                setattr(self, key, CustomParameters().fromListDict(cls_object[key]))
            elif key == "adsSchedules":
                setattr(self, key, AdsSchedules().fromListDict(cls_object[key]))
            elif key == "createdAt":
                # print('created at int', type(cls_object[key]))
                # setattr(self, key, cls_object[key]) 
                setattr(self, key, DateUtils.convert_firestore_timestamp_to_str(cls_object[key])) 
            elif key == 'realBudget':
                setattr(self, f"__{key}", cls_object[key])
            elif key == 'budget':
                if 'realBudget' not in cls_object or cls_object['realBudget'] is None or self.to_budget(cls_object['realBudget'])==0 or self.to_budget(cls_object['realBudget']) > self.to_budget(cls_object[key]):
                    setattr(self, 'realBudget', self.calculate_budget(cls_object[key]))
                setattr(self, key, self.to_budget(cls_object[key]))
            else:
                setattr(self, key, cls_object[key]) 
    def to_budget(self, value):
        try:
            if isinstance(value, (int, float)):
                return value
            if float(str(value)) == int(str(value)):
                return int(value)
            return float(str(value))
        except Exception as e:
            raise ApiResponse(ResponseStatus.ERROR, StatusCode.status_500, None, Error(f"{str(e)}","INVALID_REQUEST", 1))
    def ages_to_list(self, ages: list[Ages] | None = None):
        if ages is None:
            return [];
        return Ages().toListDict(ages, None)
    def genders_to_list(self, genders: list[Genders] | None = None):
        if genders is None:
            return [];
        return Genders().toListDict(genders, None)
    def devices_to_list(self, devices: list[Devices] | None = None):
        if devices is None:
            return [];
        return Devices().toListDict(devices, None)
    def locations_to_list(self, locations: list[Location] | None = None, fields=None):
        if locations is None:
            return []
        return Location().toListDict(locations, fields)
    def schedules_to_list(self, schedules: list[AdsSchedules] | None = None, fields=None):
        if schedules is None:
            return []
        return AdsSchedules().toListDict(schedules, fields)
    def custom_parameters_to_list(self, custom_parameters: list[CustomParameters] | None = None, fields=None):
        if custom_parameters is None:
            return []
        return CustomParameters().toListDict(custom_parameters, None, fields)
    
    def list_to_ages(self, ages: list[Ages] | None = None):
        if ages is None:
            return [];
        return Ages().fromListDict(ages, Ages)
    def list_to_genders(self, genders: list[Genders] | None = None):
        if genders is None:
            return [];
        return Genders().fromListDict(genders, Genders)
    def list_to_devices(self, devices: list[Devices] | None = None):
        if devices is None:
            return [];
        return Devices().fromListDict(devices, Devices)
    def is_published(self):
        return self.id_campagne is not None and self.id_campagne > 0
    @staticmethod
    def generate_model(_key_="default_value"):
        campaign = {};
        props = BaseCampaignFieldsProps
        for k in DictUtils.get_keys(props):
            campaign[k] = props[k][_key_];
        return campaign;

    
    def to_dict(self, fields=None):
        createdAt = DateUtils.convert_firestore_timestamp_to_str(self.createdAt)
        id_campagne = self.id_campagne
        ad_group_id = self.ad_group_id
        budget_id = self.budgetId
        ad_group_id_firebase = self.ad_group_id_firebase

        if id_campagne == 0:
            id_campagne = None
        if ad_group_id == 0:
            ad_group_id = None
        if budget_id == 0:
            budget_id = None
        if ad_group_id_firebase == "":
            ad_group_id_firebase = None

        
        return DictUtils.remove_none_values({
            "id": self.id,
            "id_campagne": id_campagne,
            "name": self.name,
            "status": self.status,
            "objective": self.objective.to_dict(None),
            "adChannel": self.adChannel.to_dict(None),
            "startDate": self.startDate,
            "endDate": self.endDate,
            "startDateFormattedGoogle": self.startDateFormattedGoogle,
            "endDateFormattedGoogle": self.endDateFormattedGoogle,  
            "areaTargetedOption": self.areaTargetedOption,
            "areaExcludedOption": self.areaExcludedOption,  
            "budget": self.budget,
            "realBudget": self.__realBudget,
            "budgetId": budget_id,
            "dailyBudget": self.dailyBudget,
            "bid": self.bid,
            "strategie": self.strategie,
            "urlPromote": self.urlPromote,
            "ages": Ages().toListDict(self.ages, None),
            "genders": Genders().toListDict(self.genders, None),
            "targetedLocations": Location().toListDict(self.targetedLocations, None),
            "excludedLocations": Location().toListDict(self.excludedLocations, None),
            "devicesTargeted": Devices().toListDict(self.devicesTargeted, None),
            "devicesExcluded": Devices().toListDict(self.devicesExcluded, None),
            "deliveryMethod": self.deliveryMethod,
            "trackingTemplate": self.trackingTemplate,
            "finalUrlSuffix": self.finalUrlSuffix,
            "accountId": self.accountId,
            "type": self.type,
            "is_removed": self.is_removed,
            "isEnded": self.isEnded,
            "isComplete": self.isComplete,
            "isPayed": self.isPayed,
            "publish": self.publish,
            "publishing": self.publishing,
            "budgetEnded": self.budgetEnded,
            "ad_group_id": ad_group_id,
            "ad_group_id_firebase": ad_group_id_firebase,
            "servingStatus": self.servingStatus,
            "owner": self.owner,
            "provider": self.provider,
            "urlCustomParameters": CustomParameters().toListDict(self.urlCustomParameters, None),
            "adsSchedules": AdsSchedules().toListDict(self.adsSchedules, None),
            "clientCustomerId": self.clientCustomerId,
            "createdAt": createdAt,
            "createdBy": self.createdBy,
            "google_ads_campaign": self.google_ads_campaign,
            "api_version": self.api_version,
        })
        


import json
from dataclasses import dataclass
from email import message
from enum import Enum
from typing import Any


class ResponseStatus:
    OK = 'ok'
    ERROR = 'error'

class StatusCode:
    """
    200: OK
    201: Created
    204: No Content
    206: Partial Content
    300: Multiple Choices
    301: Moved Permanently
    302: Found
    303: See Other
    304: Not Modified
    307: Temporary Redirect
    400: Bad Request
    401: Unauthorized
    403: Forbidden
    404: Not Found
    408: Request Timeout
    500: Internal Server Error
    502: Bad Gateway
    504: Gateway Timeout
    """
    status_200 = 200
    status_201 = 201
    status_204 = 204
    status_206 = 206
    status_300 = 300
    status_301 = 301
    status_302 = 302
    status_303 = 303
    status_304 = 304
    status_307 = 307
    status_400 = 404
    status_401 = 401
    status_402 = 402
    status_403 = 403
    status_404 = 404
    status_408 = 408
    status_500 = 500
    status_502 = 502
    status_504 = 504

@dataclass
class Error:
    message: str | dict | list
    name: str
    error_code: int = 1
    values: list = None

    @staticmethod
    def from_dict(obj: Any) -> 'Error':
        _message = obj.get("message")
        _name = str(obj.get("name"))
        _error_code = "1"
        code = obj.get("error_code")
        if code is not None:
            _error_code = int(str(code))
        if obj.get("values"):
            _values = obj.get("values")
        return Error(_message, _name, _error_code, _values)

    def to_dict(self):
        msg = self.message
        if isinstance(self.message, (dict, list)):
            msg = convert_data(self.message)
        values = None
        key = "message"
        if isinstance(self.values, (dict, list)):
            values = convert_data(self.values)
        if isinstance(self.values, list):
            key = "messages"
        data = {
            key: msg,
            "name": self.name,
            "error_code": self.error_code,
        }
        if bool(values):
            data["values"] = values.copy()
        
        return data
    def to_list(self):
        return [self.to_dict()]


def convert_data(data):
    if isinstance(data, list):
        return [convert_data(item) for item in data]
    elif isinstance(data, dict):
        return {key: convert_data(value) for key, value in data.items()}
    elif getattr(data, 'to_dict', None) is not None:
        return data.to_dict()
    else:
        return data
@dataclass
class ApiResponse:
    status: str
    status_code: int
    data: Any
    error: Error | list[Error]

    @staticmethod
    def from_dict(obj: Any) -> 'ApiResponse':
        _status = str(obj.get("status"))
        _status_code = int(obj.get("status_code"))
        _data = obj.get("data")
        _error = Error.from_dict(obj.get("error"))
        return ApiResponse(_status, _status_code, _data, _error)
    
    def to_json(self):
        return self.to_dict()

    def to_dict(self, fields=None):
        data = convert_data(self.data)
        values = {
            "status": self.status,
            "status_code": self.status_code,
            "data": data,
        }
        if getattr(self, 'error') is not None and self.error is not None:
            values['error'] = convert_data(self.error)

        if 'data' in values and values['data'] is None:
            del values['data']
        if 'error' in values and values['error'] is None:
            del values['error']
        return values

    # def to_json(self):
    #     return json.loads(JsonEncoder().encode(self))



import os
from dataclasses import dataclass
from datetime import datetime
from typing import Any

from adafri.utils import DateUtils, integer
from adafri.v1.account import Account
from google.cloud.firestore_v1.document import DocumentReference

from ....utils import ApiResponse, Error, ResponseStatus, StatusCode
from ....utils.utils import (
    BaseClass,
    BaseFieldsClass,
    DictUtils,
    get_object_model_class,
    init_class_kwargs,
)
from ...base.firebase_collection import FirebaseCollectionBase, getTimestamp
from .client_customer_id_fields import (
    CLIENT_CUSTOMER_ID_PICKABLE_FIELDS,
    ClientCustomerIdFields,
    ClientCustomerIdFieldsProps,
)

COLLECTION_NAME = os.environ.get('GADS_CLIENT_CUSTOMER_ID_COLLECTION', 'clients-customer-id');
@dataclass(init=False)
class ClientCustomerId(FirebaseCollectionBase):
    id: str
    accountId: str
    owner: str
    customerId: str
    isDefault: bool
    name: str
    enabled: bool
    createdAt: Any
    creationDate: int
    creationDateString: str
    def __init__(self, customer=None, collection_name=COLLECTION_NAME, **kwargs):
        if type(customer) == str:
            customer = {'customerId': customer}
        (cls_object, keys, data_args) = init_class_kwargs(self, customer, CLIENT_CUSTOMER_ID_PICKABLE_FIELDS, ClientCustomerIdFieldsProps, collection_name, ['customerId'], **kwargs)
        super().__init__(**data_args);
        for key in keys:
            setattr(self, key, cls_object[key]) 
    def to_dict(self, field=None):
        createdAt = None
        if getattr(self, 'createdAt', None) is not None: 
            createdAt = DateUtils.convert_firestore_timestamp_to_str(self.createdAt)
        return DictUtils.remove_none_values({
            "accountId": self.accountId,
            "customerId": self.customerId,
            "name": self.name,
            "enabled": self.enabled,
            "isDefault": self.isDefault,
            "createdAt": createdAt,
            "creationDate": self.creationDate,
            "creationDateString": self.creationDateString
        })
    
    @staticmethod
    def generate_model(_key_="default_value"):
        campaign = {};
        props = ClientCustomerIdFieldsProps
        for k in DictUtils.get_keys(props):
            campaign[k] = props[k][_key_];
        return campaign;
    # def query(self, query_params: list, first=False, limit=None):
    #     result = [];
    #     query_result = self.custom_query(query_params, first=first, limit=limit, collection_name=self.collection_name)
    #     # print('query result', query_result)
    #     if bool(query_result):
    #         if first:
    #             return ClientCustomerId.from_dict(customer=query_result, db=self.db, collection_name=self.collection_name)
    #         else:
    #             for doc in query_result:
    #                 result.append(ClientCustomerId.from_dict(customer=doc, db=self.db, collection_name=self.collection_name))
    #             return result
    #     if first:
    #             return None
    #     return [];
    def get(self):
        if getattr(self, 'customerId', None) is None:
            return None;
        query = self.query(self, 'customer', [{"key": "customerId", "comp": "==", "value": self.customerId}], True)
        if query is None:
            return None;
        result: ClientCustomerId = query
        return result;
    @staticmethod
    def from_dict(customer: Any = None, db=None, collection_name=COLLECTION_NAME) -> 'ClientCustomerId':
        cls_object, keys = get_object_model_class(customer, ClientCustomerId, ClientCustomerIdFieldsProps);
        return ClientCustomerId(cls_object, db=db, collection_name=collection_name)
    
    def getDefault(self)->'list[ClientCustomerId]':
        query = self.query(self, 'customer', [{"key": "isDefault", "comp": "==", "value": True}])
        if query is None:
            return [];
        result: list[ClientCustomerId] = query
        return result;
    def create(self, accountId, **kwargs):
        print('kwargs', kwargs)
        data_dict = DictUtils.pick_fields(kwargs, ClientCustomerIdFields(fields_props=ClientCustomerIdFieldsProps).filtered_keys('mutable', True));
        # print('data_dict', data_dict)
        customer_model = ClientCustomerId().from_dict(DictUtils.merge_dict(data_dict, ClientCustomerId().generate_model(), True));
        if bool(customer_model.to_json()) is False:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error("Empty request","INVALID_REQUEST", 1))
        print('customer_model', customer_model.customerId)
        if bool(customer_model.customerId) is False or len(customer_model.customerId) != 10 or integer(customer_model.customerId) is None:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error("Invalid or missing customerId","INVALID_REQUEST", 1));
        
        account = None
        if accountId is not None:
            account = Account(accountId).get();
        if account is None:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error("Account not found","INVALID_REQUEST", 1));
    
        customer_model.accountId = account.id;
        customer_model.owner = account.owner
        customerId = customer_model.customerId

        docs = ClientCustomerId().query(self, "customer", query_params=[{"key": "accountId", "comp": "==", "value": customer_model.accountId}, {"key": "customerId", "comp": "==", "value": customerId}]);
        found_customer = [];
        if len(docs) > 0:
            for doc in docs:
                found_customer.append(ClientCustomerId.from_dict(doc.to_dict()).to_json());
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"Customer with id {customer_model.customerId} already exist","INVALID_REQUEST", 1));
    
        try:
            doc: DocumentReference = ClientCustomerId().collection().document()
            customer_model.id = doc.id;
            if bool(customer_model.isDefault) is False:
                customer_model.isDefault = False
            creationDate = datetime.now()
            creationDateN = int(creationDate.timestamp()) * 1000
            creationDateString = creationDate.isoformat();
            creationDateString = creationDate.strftime("%d %B %Y à %H:%M:%S UTC");
            customer_model.creationDate = creationDateN;
            customer_model.creationDateString = creationDateString;
            doc.set({**customer_model.to_json(), "createdAt": getTimestamp()}, merge=True)
            return ApiResponse(ResponseStatus.OK, StatusCode.status_200, DictUtils.remove_none_values(DictUtils.filter_by_fields(customer_model.to_json(), CLIENT_CUSTOMER_ID_PICKABLE_FIELDS)), None);
        except Exception as e:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(str(e),"INVALID_REQUEST", 1));



from dataclasses import dataclass
from typing import TypeVar

from adafri.cache import Cache
from adafri.utils import BaseClass

cache = Cache(max_size=1000, eviction_policy='lru', default_ttl=3600, enable_persistence=True, persistence_file='lib_cache_data.pkl')


ClassContructor = TypeVar('ClassContructor', bound='BaseClass')

@dataclass(init=False)
class CacheClass:
    def __init__(self, prefix: str, class_constructor: ClassContructor):
        self.class_constructor = class_constructor
        self.prefix = prefix
    def set(self, key, value, ttl=3600, prefix=None):
        if value is None:
            return
        cache_key = f"{self.prefix if prefix is None else prefix}:{key}"
        if cache.get(cache_key) is None:
            if getattr(value, 'to_dict', None) is not None:
                cache.set(cache_key, value.to_dict(), ttl=ttl)
            else:
                cache.set(cache_key, value, ttl=ttl)
        
    def get(self, key, prefix=None):
        cache_key = f"{self.prefix if prefix is None else prefix}:{key}"
        data_value = cache.get(cache_key)
        if data_value is not None:
            if isinstance(data_value, dict):
                if self.class_constructor is not None:
                    return self.class_constructor.from_dict(data_value)
            return data_value
        return None



from .adsChannel import ads_channel_data
from .adschedules import days_data, times_data
from .bids import bids_data
from .demographics import ages_data, genders_data
from .devices import devices_data
from .display_ads import DISPLAY_ADS_REQUIREMENTS
from .native_ads import NATIVE_ADS_REQUIREMENTS
from .objectives import objectives_data



import base64
import hashlib
import io
import os
import secrets
from datetime import datetime, timedelta, timezone

import pyotp

from ..models.mfa import MFACredential
from ..models.mfa_fields import MFAFields


def _get_fernet():
    key = os.environ.get('MFA_ENCRYPTION_KEY')
    if key is None:
        raise ValueError('MFA_ENCRYPTION_KEY environment variable is not set')
    from cryptography.fernet import Fernet
    return Fernet(key.encode() if isinstance(key, str) else key)


def _encrypt_secret(secret: str) -> str:
    f = _get_fernet()
    return f.encrypt(secret.encode()).decode()


def _decrypt_secret(encrypted: str) -> str:
    f = _get_fernet()
    return f.decrypt(encrypted.encode()).decode()


def _hash_backup_code(code: str) -> str:
    return hashlib.sha256(code.encode()).hexdigest()


class MFAService:

    def generate_totp_enrollment(self, user_uid: str, issuer: str = None) -> dict:
        """
        Generate a TOTP enrollment secret (stored temporarily in Firestore).
        Returns {'uri': ..., 'qr_b64': ..., 'secret': ...}
        """
        if issuer is None:
            issuer = os.environ.get('OAUTH2_ISSUER', 'Adafri')
        secret = pyotp.random_base32()
        encrypted_secret = _encrypt_secret(secret)

        # Store as pending enrollment doc (10-min TTL handled by caller or Firestore TTL policy)
        from ....base.firebase_collection import getTimestamp
        from ..models.mfa import MFACredential
        pending_ref = MFACredential().collection('mfa_pending_enrollment').document(user_uid)
        expires_at = (datetime.now(timezone.utc) + timedelta(minutes=10)).isoformat()
        pending_ref.set({
            'user_uid': user_uid,
            'totp_secret': encrypted_secret,
            'expires_at': expires_at,
            'createdAt': getTimestamp(),
        })

        totp = pyotp.TOTP(secret)
        uri = totp.provisioning_uri(name=user_uid, issuer_name=issuer)

        try:
            import qrcode
            qr = qrcode.QRCode(box_size=6, border=2)
            qr.add_data(uri)
            qr.make(fit=True)
            img = qr.make_image(fill_color='black', back_color='white')
            buf = io.BytesIO()
            img.save(buf, format='PNG')
            qr_b64 = base64.b64encode(buf.getvalue()).decode()
        except ImportError:
            qr_b64 = ''

        return {'uri': uri, 'qr_b64': qr_b64, 'secret': secret}

    def confirm_enrollment(self, user_uid: str, code: str) -> list:
        """
        Verify TOTP code against pending enrollment secret, save MFACredential.
        Returns list of 10 plaintext backup codes (hashed copies stored in Firestore).
        """
        pending_ref = MFACredential().collection('mfa_pending_enrollment').document(user_uid)
        pending_doc = pending_ref.get()
        if not pending_doc.exists:
            raise ValueError('No pending enrollment found for this user')

        pending = pending_doc.to_dict()
        expires_at = pending.get('expires_at', '')
        if expires_at:
            try:
                if datetime.fromisoformat(expires_at) < datetime.now(timezone.utc):
                    raise ValueError('Enrollment expired — please restart enrollment')
            except ValueError:
                pass

        encrypted_secret = pending.get('totp_secret', '')
        secret = _decrypt_secret(encrypted_secret)

        totp = pyotp.TOTP(secret)
        if not totp.verify(code, valid_window=1):
            raise ValueError('Invalid TOTP code')

        # Generate 10 backup codes
        raw_codes = [secrets.token_hex(4).upper() + '-' + secrets.token_hex(4).upper() for _ in range(10)]
        hashed_codes = [_hash_backup_code(c) for c in raw_codes]

        # Save credential
        cred = MFACredential()
        existing = cred.get_by_user(user_uid)
        if existing:
            existing.document_reference().set({
                'totp_secret': encrypted_secret,
                'backup_codes': hashed_codes,
                'is_enabled': True,
                'enrolled_at': datetime.now(timezone.utc).isoformat(),
                'method': 'totp',
            }, merge=True)
        else:
            cred.create(
                user_uid=user_uid,
                totp_secret=encrypted_secret,
                backup_codes=hashed_codes,
                is_enabled=True,
                method='totp',
            )

        # Clean up pending doc
        pending_ref.delete()
        return raw_codes

    def verify_totp(self, user_uid: str, code: str) -> bool:
        """Verify a TOTP code with replay protection."""
        cred = MFACredential().get_by_user(user_uid)
        if cred is None or not cred.is_enabled:
            return False
        try:
            secret = _decrypt_secret(cred.totp_secret)
        except Exception:
            return False

        totp = pyotp.TOTP(secret)
        if not totp.verify(code, valid_window=1):
            return False

        # Replay protection: check if this code was already used
        used_ref = MFACredential().collection('mfa_used_codes').document(f"{user_uid}_{code}")
        used_doc = used_ref.get()
        if used_doc.exists:
            return False

        # Mark code as used (expire after 90 seconds — 1 TOTP window)
        from ....base.firebase_collection import getTimestamp
        used_ref.set({
            'user_uid': user_uid,
            'code': code,
            'used_at': datetime.now(timezone.utc).isoformat(),
            'createdAt': getTimestamp(),
        })
        # Update last_used_at
        cred.document_reference().set({'last_used_at': datetime.now(timezone.utc).isoformat()}, merge=True)
        return True

    def verify_backup_code(self, user_uid: str, code: str) -> bool:
        """Verify and consume a backup code."""
        cred = MFACredential().get_by_user(user_uid)
        if cred is None or not cred.is_enabled:
            return False

        hashed = _hash_backup_code(code.strip().upper())
        backup_codes = list(cred.backup_codes or [])
        if hashed not in backup_codes:
            return False

        backup_codes.remove(hashed)
        cred.document_reference().set({'backup_codes': backup_codes}, merge=True)
        return True

    def disable_mfa(self, user_uid: str):
        """Disable MFA for a user."""
        cred = MFACredential().get_by_user(user_uid)
        if cred is None:
            return
        cred.document_reference().set({'is_enabled': False}, merge=True)

    def is_enabled(self, user_uid: str) -> bool:
        """Return True if MFA is enabled for this user."""
        cred = MFACredential().get_by_user(user_uid)
        return cred is not None and bool(getattr(cred, 'is_enabled', False))


mfa_service = MFAService()



from .oidc_server import (
    config_oidc_oauth,
    guard_account,
    guards_oidc,
    is_admin,
    oidc_authorization_server,
    require_oidc_oauth,
)
from .server import authorization_server, config_oauth, require_oauth
from .jwks import JWKSManager, jwks_manager
from .mfa_service import MFAService, mfa_service
from .audit_service import log_event



import hashlib
import json
from dataclasses import dataclass
from typing import Any, List

import pydash
from adafri.cache.lib import CacheClass
from adafri.utils import DateUtils
from firebase_admin import auth as admin_auth
from firebase_admin.firestore import firestore

from ....utils import (
    BaseClass,
    Crypto,
    DictUtils,
    get_object_model_class,
    get_request_fields,
    init_class_kwargs,
    integer,
)
from ....utils.response import ApiResponse, Error, ResponseStatus, StatusCode
from ...base.firebase_collection import FirebaseCollectionBase, getTimestamp
from .file_fields import FILE_FIELDS, FileFieldProps, FileFields


def filter_file_request_fields(fields, default_fields = FILE_FIELDS):
    request_fields = get_request_fields(fields, default_fields, [default_fields[0]])
    if request_fields is None or bool(request_fields) is False:
        request_fields = [default_fields[0]]
    return request_fields

@dataclass(init=False)
class File(FirebaseCollectionBase):
    _id: str
    id: str
    accountId: str
    blob_id: str
    blob_name: str
    bucket_name: str
    colors: str
    content_type: str
    created_time: any
    directory_path: str
    extension: str
    file_source: str
    file_url: str
    filename: str
    height: int
    mime_type: str
    mode: str
    size: int
    width: int
    type: str
    url: str
    crop: dict
    data: str
    hash: str
    crop_state: str

    def __init__(self, file=None, **kwargs):
        value = file;
        if type(file) is str:
            value = {"id": file}
        (cls_object, keys, data_args) = init_class_kwargs(self, value, FILE_FIELDS, FileFieldProps, None, ['id','uid'], **kwargs)

        super().__init__(**data_args);
        for key in keys:
            if key in cls_object:
                if key == "created_time":
                    setattr(self, key, DateUtils.convert_firestore_timestamp_to_str(cls_object[key]))
                elif key == "width":
                    setattr(self, key, integer(cls_object[key], 0, True))
                elif key == "height":
                    setattr(self, key, integer(cls_object[key], 0, True))
                elif key == "size":
                    setattr(self, key, integer(cls_object[key], 0, True))
                else:
                    # print('set attr for', key)
                    setattr(self, key, cls_object[key]);
            else:
                setattr(self, key, None)
        if bool(self.blob_name) is True:
            col = self.blob_name.split('/');
            self.collection_name = col[0]
        if bool(self.hash) is False:
            self.hash = self.id;
        self.cache = CacheClass("file", File);
                
    


    @staticmethod
    def generate_model(_key_="default_value"):
        file = {};
        props = FileFieldProps
        for k in DictUtils.get_keys(props):
            file[k] = props[k][_key_];
        return file;

    @staticmethod
    def from_dict(file: Any=None, db=None, collection_name=None) -> 'File':
        cls_object, keys = get_object_model_class(file, File, FileFieldProps);
            
        _file = File(cls_object, db=db, collection_name=collection_name)
        return _file
    
    def to_dict(self, fields=None):
        values = {
            "id": self.id,
            "accountId": self.accountId,
            "blob_id": self.blob_id,
            "blob_name": self.blob_name,
            "bucket_name": self.bucket_name,
            "colors": self.colors,
            "content_type": self.content_type,
            "created_time": self.created_time,
            "directory_path": self.directory_path,
            "extension": self.extension,
            "file_source": self.file_source,
            "file_url": self.file_url,
            "filename": self.filename,
            "height": self.height,
            "mime_type": self.mime_type,
            "mode": self.mode,
            "size": self.size,
            "hash": self.hash,
            "width": self.width,
            "type": self.type,
            "url": self.url,
            "crop": self.crop,
            "crop_state": self.crop_state
        };
        return DictUtils.remove_none_values(values)
    def to_json(self, _fields=None):
        return self.to_dict(_fields)

    def getFiles(self, aacid, fields=None):
        files = []
        files_ = self.query(query_params=[{"key": "accountId", "comp": "==", "value": aacid}])
        request_fields = filter_file_request_fields(fields)
        # print('account query', account)
        for file in files_:
            files.append(file)
            # campaigns.append(D(campaign.to_dict(request_fields)))

        return files
    def get(self):
        file_cache: File | None = self.cache.get(self.id)
        if bool(file_cache) is True:
            return File.from_dict(file=file_cache.to_dict(), db=self.db, collection_name=self.collection_name)
        doc = self.document_reference().get();
        if doc.exists is False:
            print('file not found', self.id)
            return None;
        fields = FILE_FIELDS
        # d = pydash.pick(doc.to_dict(), fields)
        new_file = File.from_dict(file=doc.to_dict(), db=self.db, collection_name=self.collection_name)
        ttl = 225
        if new_file.crop_state == "done":
            ttl = 3600
        self.cache.set(self.id, new_file.to_dict(), ttl)
        return new_file;

    def query(self, query_params: list, first=False, limit=None):
        hash = f"{hashlib.sha1(str(str(json.dumps(query_params))).encode()).hexdigest()}"
        cache_get = self.cache.get(hash, "file_query")
        result = [];
        if bool(cache_get) is True:
            print('from cache')
            if first:
                if isinstance(cache_get, list):
                    return File.from_dict(file=cache_get[0], db=self.db, collection_name=self.collection_name)
                return File.from_dict(file=cache_get, db=self.db, collection_name=self.collection_name)
            for doc in cache_get:
                print('doc', doc)
                result.append(File.from_dict(file=doc, db=self.db, collection_name=self.collection_name))
            return result
        query_result = self.custom_query(query_params, first=first, limit=limit, collection_name=self.collection_name)
        # print('query result', query_result)
        if bool(query_result):
            if first:
                value = File.from_dict(file=query_result, db=self.db, collection_name=self.collection_name)
                self.cache.set(hash, value.to_dict(), 3600, "file_query")
                return value
            else:
                for doc in query_result:
                    result.append(File.from_dict(file=doc, db=self.db, collection_name=self.collection_name))
                self.cache.set(hash, File().toListDict(result, None), 3600, "file_query")
                return result
        if first:
                return None
        return [];

    def create(self, **kwargs):
        testMode = kwargs.get('testMode', False)
        data_dict = DictUtils.pick_fields(kwargs, FileFields(fields_props=FileFieldProps).filtered_keys('mutable', True));
        # print('file dict', DictUtils.merge_dict(data_dict, File.generate_model()))
        file_model = File().from_dict(DictUtils.merge_dict(data_dict, File.generate_model()));
        # print('file model', file_model)
        if bool(file_model.to_dict()) is False:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error("Empty request","INVALID_REQUEST", 1))
        document_id = file_model._id;
        doc_ref = self.collection().document(document_id);
        create = {**file_model.to_json(), 'createdAt': firestore.SERVER_TIMESTAMP}
        try:
            if bool(testMode) is False:
                doc_ref.set(create, True);
            data_file = File.from_dict(create)
            return ApiResponse(ResponseStatus.OK, StatusCode.status_200, data_file, None);
        
        except Exception as e:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(str(e),"INVALID_REQUEST", 1));
    
    def update(self, data):
        try:
            last_value = self.to_dict();
            filtered_value = pydash.pick(data, FileFields(fields_props=FileFieldProps).filtered_keys('editable', True));
            new_value = DictUtils.merge_dict(filtered_value, self.to_dict());
            changed_fields = DictUtils.get_changed_field(last_value, new_value);
            # print('last', last_value)
            # print('new', new_value)
            # print('changed', changed_fields)
            # print('data', data)
            data_update = DictUtils.dict_from_keys(filtered_value, changed_fields);
            if bool(data_update) is False:
                return None;
            self.document_reference().set(data_update, merge=True)
            new_file = self.get()
            self.cache.set(new_file.id, new_file)
            return DictUtils.dict_from_keys(new_file.to_dict(), changed_fields);
        except Exception as e:
            print(e)
            return None;

    def remove(self, only_mark_as_removed=True):
        try:
            if self.id is None:
                return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"Cannot identify File with id {self.id}","INVALID_REQUEST", 1));
            if only_mark_as_removed:
                self.document_reference().set({"is_removed": True})
            else:
                self.document_reference().delete();
            return ApiResponse(ResponseStatus.OK, StatusCode.status_200, {"message": f"File {self.id} deleted"}, None);
        except:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"An error occurated while removing file with id {self.id}","INVALID_REQUEST", 1));
  





from dataclasses import dataclass
from datetime import datetime
from typing import Any

from adafri.utils import DateUtils
from google.cloud.firestore_v1.transforms import Sentinel

from ......utils import DictUtils, get_request_fields, init_class_kwargs
from .....base.firebase_collection import FirebaseCollectionBase
from .base_ad_fields import BASE_AD_PICKABLE_FIELDS, STANDARD_FIELDS, BaseAdFieldsProps


def filter_request_fields(fields, default_fields = BASE_AD_PICKABLE_FIELDS):
    request_fields = get_request_fields(fields, default_fields, [default_fields[0]])
    if request_fields is None or bool(request_fields) is False:
        request_fields = [default_fields[0]]
    return request_fields



@dataclass(init=False)
class BaseAd(FirebaseCollectionBase):
    id: str
    ad_id: int
    ad_group_id: int
    adgroup_id_parent: str
    name: str
    status: str
    is_removed: bool = False
    owner: str
    clientCustomerId: str
    accountId: str
    createdAt: any
    createdBy: str
    class_object = None
    api_version: str = "v0"
    __baseFields: BaseAdFieldsProps

    def __init__(self, ad=None, collection_name=None, **kwargs):
        _ad = ad
        if type(ad) is str:
            _ad = {"id": ad}
        # if ad is not None:
        #     print('init first', ad['id'])
        (cls_object, keys, data_args) = init_class_kwargs(self, _ad, STANDARD_FIELDS, BaseAdFieldsProps, collection_name, ['id'], **kwargs)
        super().__init__(**data_args);
        for key in keys:
            if key == "createdAt":
                createdAt = DateUtils.convert_firestore_timestamp_to_str(cls_object[key]);
                if isinstance(createdAt, Sentinel) is True:
                    createdAt = datetime.now().strftime('%Y-%m-%d %H:%M:%S');
                setattr(self, key, createdAt) 
            else:
                setattr(self, key, cls_object[key]) 
        self.class_object = BaseAd

    def is_published(self):
        return self.ad_id is not None and self.ad_id > 0


    @staticmethod
    def generate_model(_key_="default_value"):
        ad = {};
        props = BaseAdFieldsProps
        for k in DictUtils.get_keys(props):
            ad[k] = props[k][_key_];
        return ad;

    
    def to_dict(self, fields=None):
        ad_id = self.ad_id
        ad_group_id = self.ad_group_id
        if ad_id == 0:
            ad_id = None;
        if ad_group_id == 0:
            ad_group_id = None;
        
        return DictUtils.remove_none_values({
            "id": self.id,
            "ad_id": ad_id,
            "name": self.name,
            "status": self.status,
            "is_removed": self.is_removed,
            "ad_group_id": ad_group_id,
            "owner": self.owner,
            "clientCustomerId": self.clientCustomerId,
            "accountId": self.accountId,
            "adgroup_id_parent": self.adgroup_id_parent,
            "createdAt": DateUtils.convert_firestore_timestamp_to_str(self.createdAt),
            "createdBy": self.createdBy,
            "api_version": self.api_version
        })
        


from dataclasses import dataclass
from  .....utils.utils import DictUtils, BaseFieldsClass 
from ...models.base_campaign_fields import BaseCampaignFields, BaseCampaignFieldsProps
import os

DISPLAY_COLLECTION = os.environ.get('GADS_DISPLAY_COLLECTION');

@dataclass
class CampaignDisplayFields(BaseCampaignFields):
    targetedPlacements = "targetedPlacements"
    excludedPlacements = "excludedPlacements"
    websites = "websites"
    youtubeChannels = "youtubeChannels"
    youtubeVideos = "youtubeVideos"
    images = "images"
    imagesNative = "imagesNative"
    user_interest = "user_interest"
    target_google_search = "target_google_search"
    target_search_network = "target_search_network"
    target_content_network = "target_content_network"
    target_partner_search_network = "target_partner_search_network"
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.class_fields_props = CampaignDisplayFieldsProps



CampaignDisplayFieldsProps = {
    **BaseCampaignFieldsProps,
    BaseCampaignFields.type: {
        "internal": True,
        "type": str,
        "required": False,
        "mutable": False,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": "DISPLAY"
    },
    CampaignDisplayFields.targetedPlacements: {
        "internal": True,
        "type": list,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": []
    },
    CampaignDisplayFields.excludedPlacements: {
        "internal": True,
        "type": list,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": []
    },
    CampaignDisplayFields.websites: {
        "internal": True,
        "type": list,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": []
    },
    CampaignDisplayFields.youtubeChannels: {
        "internal": True,
        "type": list,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": []
    },
    CampaignDisplayFields.youtubeVideos: {
        "internal": True,
        "type": list,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": []
    },
    CampaignDisplayFields.user_interest: {
        "internal": True,
        "type": list,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": []
    },
    CampaignDisplayFields.images: {
        "internal": True,
        "type": list,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": []
    },
    CampaignDisplayFields.imagesNative: {
        "internal": True,
        "type": list,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": []
    },
    CampaignDisplayFields.target_google_search: {
        "internal": True,
        "type": bool,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": False
    },
    CampaignDisplayFields.target_search_network: {
        "internal": True,
        "type": bool,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": False
    },
    CampaignDisplayFields.target_content_network: {
        "internal": True,
        "type": bool,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": True
    },
    CampaignDisplayFields.target_partner_search_network: {
        "internal": True,
        "type": bool,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": False
    }
}


STANDARD_FIELDS = CampaignDisplayFields(fields_props=CampaignDisplayFieldsProps).filtered_keys('pickable', True)
CAMPAIGN_DISPLAY_PICKABLE_FIELDS = CampaignDisplayFields(fields_props=CampaignDisplayFieldsProps).filtered_keys('pickable', True)
CAMPAIGN_DISPLAY_NOT_PICKABLE_FIELDS = CampaignDisplayFields(fields_props=CampaignDisplayFieldsProps).filtered_keys('pickable', False)



from googleapiclient.discovery import build as discovery_build
import os

YOUTUBE_API_SERVICE_NAME = os.environ.get('YOUTUBE_API_SERVICE_NAME')
YOUTUBE_API_VERSION = os.environ.get('YOUTUBE_API_VERSION')
YOUTUBE_API_DEVELOPER_KEY = os.environ.get('YOUTUBE_API_DEVELOPER_KEY')

def search_youtube_channel_by_id(query): 
    if YOUTUBE_API_SERVICE_NAME is None or YOUTUBE_API_VERSION is None or YOUTUBE_API_DEVELOPER_KEY is None:
        return []
    youtube_object = discovery_build(YOUTUBE_API_SERVICE_NAME, YOUTUBE_API_VERSION, developerKey = YOUTUBE_API_DEVELOPER_KEY) 
    search_keyword = youtube_object.channels().list(id=query, part="id, snippet", maxResults=1).execute() 
    # extracting the results from search response 
    results = search_keyword.get("items", [])
    print(results)
    channels = [] 
    # extracting required info from each result object 
    for result in results:
        if result['kind'] == "youtube#channel":
            channels.append({
                'id': '',
                'channelId': result["id"],
                'name': result["snippet"]["title"],
                'snippet': result['snippet']['description'],
                'thumbnails': result['snippet']['thumbnails']['medium']['url'],
            })
    return channels

def search_youtube_video_by_id(query): 
    if YOUTUBE_API_SERVICE_NAME is None or YOUTUBE_API_VERSION is None or YOUTUBE_API_DEVELOPER_KEY is None:
        return []
    youtube_object = discovery_build(YOUTUBE_API_SERVICE_NAME, YOUTUBE_API_VERSION, developerKey = YOUTUBE_API_DEVELOPER_KEY) 
    search_keyword = youtube_object.videos().list(id=query, part="id, snippet", maxResults=1).execute() 
    # extracting the results from search response 
    results = search_keyword.get("items", [])  
    videos = []
    print(results)
    # extracting required info from each result object 
    for result in results:
        if result['kind'] == "youtube#video":
            videos.append({
                'id': '',
                'videoId': result["id"],
                'name': result["snippet"]["title"],
                'snippet': result['snippet']['description'],
                'thumbnails': result['snippet']['thumbnails']['medium']['url'],
            })
    return videos



import uuid
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Any, Optional

from adafri.v1.base.firebase_collection import FirebaseCollectionBase

INVITE_COLLECTION = 'workflow_invites'

INVITE_PICKABLE_FIELDS = [
    'id', 'workspaceId', 'email', 'role', 'teamIds',
    'invitedBy', 'createdAt', 'status',
]

INVITE_MUTABLE_FIELDS = ['status', 'teamIds']

INVITE_STATUSES = ['pending', 'accepted', 'revoked']


def _now() -> str:
    return datetime.now(timezone.utc).isoformat()


def _new_id() -> str:
    return str(uuid.uuid4())


@dataclass(init=False)
class WorkflowInvite(FirebaseCollectionBase):
    id: str
    workspaceId: str
    email: str
    role: str
    teamIds: list
    invitedBy: str
    createdAt: str
    status: str

    def __init__(self, data=None, **kwargs):
        super().__init__(collection_name=INVITE_COLLECTION, **kwargs)
        d = data or {}
        self.id = d.get('id') or _new_id()
        self.workspaceId = d.get('workspaceId', '')
        self.email = d.get('email', '')
        self.role = d.get('role', 'viewer')
        self.teamIds = list(d.get('teamIds') or [])
        self.invitedBy = d.get('invitedBy', '')
        self.createdAt = d.get('createdAt') or _now()
        self.status = d.get('status', 'pending')

    @classmethod
    def from_dict(cls, obj: Any, **kwargs) -> 'WorkflowInvite':
        return cls(obj, **kwargs)

    def to_dict(self) -> dict:
        return {
            'id': self.id,
            'workspaceId': self.workspaceId,
            'email': self.email,
            'role': self.role,
            'teamIds': self.teamIds,
            'invitedBy': self.invitedBy,
            'createdAt': self.createdAt,
            'status': self.status,
        }

    @staticmethod
    def toListDict(items: list, fields=None) -> list:
        return [item.to_dict() for item in items]

    @classmethod
    def find(cls, invite_id: str) -> 'Optional[WorkflowInvite]':
        inst = cls()
        doc = inst.collection().document(invite_id).get()
        if not doc.exists:
            return None
        return cls({**doc.to_dict(), 'id': doc.id})

    @classmethod
    def findByEmail(cls, workspace_id: str, email: str) -> 'Optional[WorkflowInvite]':
        inst = cls()
        docs = (
            inst.collection()
            .where('workspaceId', '==', workspace_id)
            .where('email', '==', email)
            .where('status', '==', 'pending')
            .limit(1)
            .stream()
        )
        for d in docs:
            return cls({**d.to_dict(), 'id': d.id})
        return None

    @classmethod
    def findByWorkspace(cls, workspace_id: str) -> 'list[WorkflowInvite]':
        inst = cls()
        docs = (
            inst.collection()
            .where('workspaceId', '==', workspace_id)
            .order_by('createdAt')
            .stream()
        )
        return [cls({**d.to_dict(), 'id': d.id}) for d in docs]

    def save(self) -> 'WorkflowInvite':
        self.collection().document(self.id).set(self.to_dict())
        return self

    def update(self, data: dict) -> 'WorkflowInvite':
        filtered = {k: v for k, v in data.items() if k in INVITE_MUTABLE_FIELDS}
        self.collection().document(self.id).update(filtered)
        for k, v in filtered.items():
            setattr(self, k, v)
        return self

    def delete(self) -> None:
        self.collection().document(self.id).delete()



from ..models.audit_log import AuditLog


def log_event(
    event_type: str,
    request=None,
    user_uid: str = "",
    client_id: str = "",
    metadata: dict = None,
):
    """
    Log an audit event to Firestore.

    Supported event_types:
        signin_success, signin_failure, signup_success,
        mfa_verified, mfa_failed, mfa_enrolled,
        token_issued, token_revoked, token_reuse_detected,
        session_created, session_revoked,
        device_authorized, client_registered
    """
    ip_address = ""
    user_agent = ""

    if request is not None:
        try:
            ip_address = request.environ.get('HTTP_X_FORWARDED_FOR', request.remote_addr or '')
            user_agent = request.headers.get('User-Agent', '')
        except Exception:
            pass

    try:
        AuditLog.log(
            event_type=event_type,
            user_uid=user_uid or "",
            client_id=client_id or "",
            ip_address=ip_address,
            user_agent=user_agent,
            metadata=metadata or {},
        )
    except Exception as e:
        print(f'audit_service: failed to log event {event_type}: {e}')



from .client_fields import (ClientFieldProps, ClientFields)
from .client import (OAuthClient)
from .mfa import MFACredential
from .mfa_fields import MFAFields, MFAFieldsProps
from .login_session import LoginSession
from .login_session_fields import LoginSessionFields, LoginSessionFieldsProps
from .audit_log import AuditLog
from .audit_log_fields import AuditLogFields, AuditLogFieldsProps
from .device_code import DeviceCode
from .device_code_fields import DeviceCodeFields, DeviceCodeFieldsProps
from .token import TokenIntrospectionEndpoint


from dataclasses import dataclass
from .....utils import DictUtils
import os

AUDIT_LOG_COLLECTION = os.environ.get('AUDIT_LOG_COLLECTION', 'audit_logs')


@dataclass
class AuditLogFields:
    id = "id"
    event_type = "event_type"
    user_uid = "user_uid"
    client_id = "client_id"
    ip_address = "ip_address"
    user_agent = "user_agent"
    metadata = "metadata"
    created_at = "created_at"

    @staticmethod
    def keys():
        return DictUtils.get_keys(AuditLogFieldsProps)

    @staticmethod
    def filtered_keys(field, condition=True):
        mutable = DictUtils.filter(AuditLogFieldsProps, DictUtils.get_keys(AuditLogFieldsProps), field, condition)
        return DictUtils.get_keys(mutable)


AuditLogFieldsProps = {
    AuditLogFields.id: {
        "type": str,
        "required": True,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
    AuditLogFields.event_type: {
        "type": str,
        "required": True,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
    AuditLogFields.user_uid: {
        "type": str,
        "required": False,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
    AuditLogFields.client_id: {
        "type": str,
        "required": False,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
    AuditLogFields.ip_address: {
        "type": str,
        "required": False,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
    AuditLogFields.user_agent: {
        "type": str,
        "required": False,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
    AuditLogFields.metadata: {
        "type": dict,
        "required": False,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "default_value": {},
        "pickable": True
    },
    AuditLogFields.created_at: {
        "type": str,
        "required": False,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "default_value": "",
        "pickable": True
    },
}

STANDARD_FIELDS = AuditLogFields.filtered_keys('pickable', True)
